"""
Encoder and Decoder for PyTorch Tensor Data via Wrapyfi.

This script provides mechanisms to encode and decode PyTorch tensor data using Wrapyfi.
It utilizes base64 encoding to convert binary data into ASCII strings.

The script contains a class, `PytorchTensor`, registered as a plugin to manage the
conversion of PyTorch tensor data (if available) between its original and encoded forms.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - PyTorch: An open-source machine learning library developed by Facebook's AI Research lab (refer to https://pytorch.org/get-started/locally/ for installation instructions)
        Note: If PyTorch is not available, HAVE_TORCH will be set to False and
        the plugin will be registered with no types.

    You can install the necessary packages using pip:
        ``pip install torch``  # Basic installation of PyTorch
"""

import io
import base64

from wrapyfi.utils import *

try:
    import torch

    HAVE_TORCH = True
except ImportError:
    HAVE_TORCH = False


@PluginRegistrar.register(types=None if not HAVE_TORCH else torch.Tensor.__mro__[:-1])
class PytorchTensor(Plugin):
    def __init__(self, load_torch_device=None, map_torch_devices=None, **kwargs):
        """
        Initialize the PytorchTensor plugin.

        :param load_torch_device: str: Default PyTorch device to load tensors onto
        :param map_torch_devices: dict: A dictionary mapping encoded device strings to decoding devices
        """
        self.map_torch_devices = map_torch_devices or {}
        if load_torch_device is not None:
            self.map_torch_devices["default"] = load_torch_device

    def encode(self, obj, *args, **kwargs):
        """
        Encode PyTorch tensor data into a base64 ASCII string.

        :param obj: torch.Tensor: The PyTorch tensor data to encode
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: Always True, indicating that the encoding was successful
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name, encoded data string, and device string
        """
        with io.BytesIO() as memfile:
            torch.save(obj, memfile)
            obj_data = base64.b64encode(memfile.getvalue()).decode("ascii")
        obj_device = str(obj.device)
        return True, dict(
            __wrapyfi__=(str(self.__class__.__name__), obj_data, obj_device)
        )

    def decode(self, obj_type, obj_full, *args, **kwargs):
        """
        Decode a base64 ASCII string back into PyTorch tensor data.

        :param obj_type: type: The expected type of the decoded object (not used)
        :param obj_full: tuple: A tuple containing the encoded data string and device string
        :param args: tuple: Additional arguments (not used)
        :param kwargs: dict: Additional keyword arguments (not used)
        :return: Tuple[bool, torch.Tensor]: A tuple containing:
            - bool: Always True, indicating that the decoding was successful
            - torch.Tensor: The decoded PyTorch tensor data
        """
        with io.BytesIO(base64.b64decode(obj_full[1].encode("ascii"))) as memfile:
            obj_device = self.map_torch_devices.get(
                obj_full[2], self.map_torch_devices.get("default", None)
            )
            if obj_device is not None:
                return True, torch.load(memfile, map_location=obj_device)
            else:
                return True, torch.load(memfile)



import logging
import json
import time
import os
import queue
from typing import Optional

import numpy as np
import cv2

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog
from wrapyfi.middlewares.websocket import WebSocketMiddlewarePubSub
from wrapyfi.encoders import JsonDecodeHook

SOCKET_IP = os.environ.get("WRAPYFI_WEBSOCKET_SOCKET_IP", "127.0.0.1")
SOCKET_PORT = int(os.environ.get("WRAPYFI_WEBSOCKET_SOCKET_PORT", 5000))
WEBSOCKET_NAMESPACE = os.environ.get("WRAPYFI_WEBSOCKET_NAMESPACE", "/")
WEBSOCKET_MONITOR_LISTENER_SPAWN = os.environ.get(
    "WRAPYFI_WEBSOCKET_MONITOR_LISTENER_SPAWN", "thread"
)
WATCHDOG_POLL_REPEAT = None
if WEBSOCKET_MONITOR_LISTENER_SPAWN == "process":
    WEBSOCKET_MONITOR_LISTENER_SPAWN = "thread"
    logging.warning(
        "[WebSocket] Wrapyfi does not support multiprocessing for Websockets. Please set "
        "the environment variable WRAPYFI_WEBSOCKET_MONITOR_LISTENER_SPAWN='thread'. "
        "Switching automatically to 'thread' mode."
    )


class WebSocketListener(Listener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        socket_ip: str = SOCKET_IP,
        socket_port: int = SOCKET_PORT,
        namespace: str = WEBSOCKET_NAMESPACE,
        monitor_listener_spawn: Optional[str] = WEBSOCKET_MONITOR_LISTENER_SPAWN,
        websocket_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the subscriber.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic (e.g., 'topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param socket_ip: str: IP address of the socket. Default is '127.0.0.1'
        :param socket_port: int: Port of the socket for publishing. Default is 5000
        :param namespace: str: Namespace of the WebSocket. Default is '/'
        :param monitor_listener_spawn: str: Whether to spawn the monitor listener as a process or thread. Default is 'thread'
        :param websocket_kwargs: dict: Additional kwargs for the WebSocket middleware
        :param kwargs: dict: Additional kwargs for the subscriber
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)

        self.socket_address = f"http://{socket_ip}:{socket_port}{namespace}"

        # Activate the WebSocket middleware
        WebSocketMiddlewarePubSub.activate(
            socket_address=self.socket_address,
            monitor_listener_spawn=monitor_listener_spawn,
            **(websocket_kwargs or {}),
        )

        # Register the callback for the topic
        WebSocketMiddlewarePubSub._instance.register_callback(
            self.in_topic, self.on_message
        )

        if not self.should_wait:
            ListenerWatchDog().add_listener(self)

    def await_connection(
        self, in_topic: Optional[str] = None, repeats: Optional[int] = None
    ):
        """
        Wait until the WebSocket connection is established.

        :param in_topic: str: The topic to monitor for connection
        :param repeats: int: The number of times to check for the connection, None for infinite.
        """
        if in_topic is None:
            in_topic = self.in_topic
        logging.info(f"[WebSocket] Waiting for input port: {in_topic}")
        if repeats is None:
            repeats = -1 if self.should_wait else 0

        # Ensure to call is_connected() on the singleton instance
        while repeats > 0 or repeats == -1:
            if repeats != -1:
                repeats -= 1
            connected = (
                WebSocketMiddlewarePubSub._instance.is_connected()
            )  # Use the instance
            logging.debug(f"Connection status: {connected}")
            if connected:
                logging.info(f"[WebSocket] Connected to input port: {in_topic}")
                return True
            time.sleep(0.2)
        return False

    def close(self):
        """
        Close the subscriber.
        """
        # You might want to deregister the callback or perform other cleanup here
        # However, since the middleware handles cleanup on exit, we might not need to do anything
        pass

    def __del__(self):
        self.close()


@Listeners.register("NativeObject", "websocket")
class WebSocketNativeObjectListener(WebSocketListener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObject listener using the WebSocket message construct assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a native object.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic (e.g., 'topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}
        self._message_queue = queue.Queue()

    def on_message(self, data):
        try:
            obj = json.loads(
                data,
                object_hook=self._plugin_decoder_hook,
                **self._deserializer_kwargs,
            )
            self._message_queue.put(obj)
            logging.debug(f"Message queued for topic {self.in_topic}: {obj}")
        except json.JSONDecodeError as e:
            logging.error(f"Failed to decode JSON from topic {self.in_topic}: {e}")

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(repeats=repeats)
        return self.check_establishment(established)

    def listen(self):
        """
        Listen for a message.

        :return: Any: The received message as a native Python object
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None

        try:
            obj = self._message_queue.get(block=self.should_wait)
            return obj
        except queue.Empty:
            return None


@Listeners.register("Image", "websocket")
class WebSocketImageListener(WebSocketNativeObjectListener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        **kwargs,
    ):
        """
        The Image listener using the WebSocket message construct parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic (e.g., 'topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8
        self._message_queue = queue.Queue()

    def on_message(self, data):
        """
        Callback for handling incoming image messages.
        """
        try:
            header, img_bytes = data
            if self.jpg:
                # JPEG case: decode the JPEG image
                img_array = np.frombuffer(img_bytes, dtype=np.uint8)
                if self.rgb:
                    img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)
                else:
                    img = cv2.imdecode(img_array, cv2.IMREAD_GRAYSCALE)
                self._message_queue.put(img)
            else:
                # Non-JPEG case: reconstruct the numpy array
                shape = header.get("shape", None)
                dtype = header.get("dtype", None)
                if shape is not None and dtype is not None:
                    img_array = np.frombuffer(img_bytes, dtype=dtype)
                    img = img_array.reshape(shape)
                    self._message_queue.put(img)
                else:
                    logging.error(
                        "Missing 'shape' or 'dtype' in header for non-JPEG image"
                    )
        except Exception as e:
            logging.error(f"Failed to process message: {e}")

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        return super().establish(repeats=repeats, **kwargs)

    def listen(self):
        """
        Listen for a message.

        :return: np.ndarray: The received image as a numpy array formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None

        try:
            img = self._message_queue.get(block=self.should_wait)
            if (
                (self.width > 0 and self.width != img.shape[1])
                or (self.height > 0 and self.height != img.shape[0])
                or not (
                    (img.ndim == 2 and not self.rgb)
                    or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
                )
            ):
                raise ValueError("Incorrect image shape for listener")
            return img
        except queue.Empty:
            return None


@Listeners.register("AudioChunk", "websocket")
class WebSocketAudioChunkListener(WebSocketNativeObjectListener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        **kwargs,
    ):
        """
        The AudioChunk listener using the WebSocket message construct parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic (e.g., 'topic')
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._message_queue = queue.Queue()

    def on_message(self, data):
        """
        Callback for handling incoming audio messages.
        """
        try:
            header, aud_bytes = data
            shape = header.get("shape")
            dtype = header.get("dtype")
            rate = header.get("rate")
            timestamp = header.get("timestamp")

            if shape is None or dtype is None or rate is None:
                raise ValueError("Missing 'shape', 'dtype', or 'rate' in header")

            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for listener")

            # Reconstruct the audio array from the binary data
            aud_array = np.frombuffer(aud_bytes, dtype=dtype).reshape(shape)

            chunk, channels = (
                aud_array.shape if len(aud_array.shape) > 1 else (aud_array.shape[0], 1)
            )

            if (0 < self.chunk != chunk) or (0 < self.channels != channels):
                raise ValueError("Incorrect audio shape for listener")

            self._message_queue.put((aud_array, rate))
        except Exception as e:
            logging.error(f"Failed to process message: {e}")

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        return super().establish(repeats=repeats, **kwargs)

    def listen(self):
        """
        Listen for a message.

        :return: Tuple[np.ndarray, int]: The received audio chunk as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None, self.rate

        try:
            aud, rate = self._message_queue.get(block=self.should_wait)
            return aud, rate
        except queue.Empty:
            return None, self.rate


@Listeners.register("Properties", "websocket")
class WebSocketPropertiesListener(WebSocketNativeObjectListener):
    def __init__(self, name, in_topic, **kwargs):
        super().__init__(name, in_topic, **kwargs)
        raise NotImplementedError



import logging
import json
import time
import os
import threading
from typing import Optional, Tuple
import random
import struct

import numpy as np
import cv2

from wrapyfi.connect.publishers import Publisher, Publishers, PublisherWatchDog
from wrapyfi.middlewares.mqtt import MqttMiddlewarePubSub
from wrapyfi.encoders import JsonEncoder

MQTT_BROKER_ADDRESS = os.environ.get("WRAPYFI_MQTT_BROKER_ADDRESS", "broker.emqx.io")
MQTT_BROKER_PORT = int(os.environ.get("WRAPYFI_MQTT_BROKER_PORT", 1883))


class MqttPublisher(Publisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        broker_address: str = MQTT_BROKER_ADDRESS,
        broker_port: int = MQTT_BROKER_PORT,
        mqtt_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the publisher.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param broker_address: str: Address of the MQTT broker. Default is 'broker.emqx.io'
        :param broker_port: int: Port of the MQTT broker. Default is 1883
        :param mqtt_kwargs: dict: Additional kwargs for the MQTT middleware
        :param kwargs: dict: Additional kwargs for the publisher
        """
        super().__init__(name, out_topic, should_wait=should_wait, **kwargs)

        if mqtt_kwargs is None:
            mqtt_kwargs = {}

        if "client_id" not in mqtt_kwargs:
            mqtt_kwargs["client_id"] = f"client-{name}-{random.randint(0, 100000)}"
        # if "clean_session" not in mqtt_kwargs:
        #     mqtt_kwargs["clean_session"] = False

        # Activate the MQTT middleware
        MqttMiddlewarePubSub.activate(
            broker_address=broker_address,
            broker_port=broker_port,
            **(mqtt_kwargs or {}),
        )

    def await_connection(
        self, out_topic: Optional[str] = None, repeats: Optional[int] = None
    ):
        """
        Wait for the connection to be established.

        :param out_topic: str: Name of the output topic
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        if out_topic is None:
            out_topic = self.out_topic
        logging.info(f"[MQTT] Waiting for output connection: {out_topic}")
        if repeats is None:
            repeats = -1 if self.should_wait else 0

        while repeats > 0 or repeats == -1:
            if repeats != -1:
                repeats -= 1
            connected = MqttMiddlewarePubSub._instance.is_connected()
            if connected:
                logging.info(f"[MQTT] Output connection established: {out_topic}")
                return True
            time.sleep(0.02)
        return False

    def close(self):
        """
        Close the publisher.
        """
        logging.info(f"[MQTT] Closing publisher for topic: {self.out_topic}")
        MqttMiddlewarePubSub._instance.mqtt_client.disconnect()
        time.sleep(0.2)

    def __del__(self):
        self.close()


@Publishers.register("NativeObject", "mqtt")
class MqttNativeObjectPublisher(MqttPublisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        multi_threaded: bool = False,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObjectPublisher using the MQTT message construct assuming a combination of python native objects
        and numpy arrays as input. Serializes the data (including plugins) using the encoder and sends it as a string.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate client for each thread. Default is False
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param kwargs: dict: Additional kwargs for the publisher
        """
        super().__init__(name, out_topic, should_wait=should_wait, **kwargs)
        if multi_threaded:
            self._thread_local_storage = threading.local()

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(repeats=repeats)
        return self.check_establishment(established)

    def publish(self, obj):
        """
        Publish the object to the middleware.

        :param obj: object: Object to publish
        """
        if not self.established:
            established = self.establish(repeats=10)
            if not established:
                return
            else:
                time.sleep(0.2)

        obj_str = json.dumps(
            obj,
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            **self._serializer_kwargs,
        )
        MqttMiddlewarePubSub._instance.mqtt_client.publish(self.out_topic, obj_str)


@Publishers.register("Image", "mqtt")
class MqttImagePublisher(MqttNativeObjectPublisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        multi_threaded: bool = False,
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        **kwargs,
    ):
        """
        The ImagePublisher using the MQTT message construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate client for each thread. Default is False
        :param width: int: Width of the image. Default is -1 meaning that the width is not fixed
        :param height: int: Height of the image. Default is -1 meaning that the height is not fixed
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed as JPG. Default is False
        """
        super().__init__(name, out_topic, should_wait=should_wait, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8

    def publish(self, img: np.ndarray):
        """
        Publish the image to the middleware.

        :param img: np.ndarray: Image to publish formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if img is None:
            return

        if not self.established:
            established = self.establish(repeats=10)
            if not established:
                return
            else:
                time.sleep(0.2)

        if (
            0 < self.width != img.shape[1]
            or 0 < self.height != img.shape[0]
            or not (
                (img.ndim == 2 and not self.rgb)
                or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
            )
        ):
            raise ValueError("Incorrect image shape for publisher")
        if not img.flags["C_CONTIGUOUS"]:
            img = np.ascontiguousarray(img)

        if self.jpg:
            # Encode image as JPEG and get raw bytes
            img_bytes = cv2.imencode(".jpg", img)[1].tobytes()
            header = {
                "timestamp": time.time(),
            }
        else:
            # Convert the image to raw bytes
            img_bytes = img.tobytes()
            header = {
                "shape": img.shape,
                "dtype": str(img.dtype),
                "timestamp": time.time(),
            }

        # Serialize header to JSON and encode to bytes
        header_json = json.dumps(header)
        header_bytes = header_json.encode("utf-8")
        header_length = len(header_bytes)
        # Pack header length into 4 bytes (big-endian)
        header_length_packed = struct.pack("!I", header_length)

        # Construct the payload: header length + header bytes + image bytes
        payload = header_length_packed + header_bytes + img_bytes

        # Publish the binary payload
        MqttMiddlewarePubSub._instance.mqtt_client.publish(self.out_topic, payload)


@Publishers.register("AudioChunk", "mqtt")
class MqttAudioChunkPublisher(MqttNativeObjectPublisher):
    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        multi_threaded: bool = False,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        **kwargs,
    ):
        """
        The AudioChunkPublisher using the MQTT message construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate client for each thread. Default is False
        :param channels: int: Number of channels. Default is 1
        :param rate: int: Sampling rate. Default is 44100
        :param chunk: int: Chunk size. Default is -1 meaning that the chunk size is not fixed
        """
        super().__init__(
            name,
            out_topic,
            should_wait=should_wait,
            multi_threaded=multi_threaded,
            **kwargs,
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    import struct
    import json
    import numpy as np
    import time
    import logging

    def publish(self, aud: Tuple[np.ndarray, int]):
        """
        Publish the audio chunk to the middleware.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=10)
            if not established:
                return
            else:
                time.sleep(0.2)

        aud_array, rate = aud
        if aud_array is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")

        chunk, channels = (
            aud_array.shape if len(aud_array.shape) > 1 else (aud_array.shape[0], 1)
        )
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")

        aud_array = np.require(aud_array, dtype=np.float32, requirements="C")

        # Create the header
        header = {
            "shape": aud_array.shape,
            "dtype": str(aud_array.dtype),
            "rate": rate,
            "timestamp": time.time(),
        }

        # Serialize header to JSON and encode to bytes
        header_json = json.dumps(header)
        header_bytes = header_json.encode("utf-8")
        header_length = len(header_bytes)
        # Pack header length into 4 bytes (big-endian)
        header_length_packed = struct.pack("!I", header_length)

        # Get audio bytes
        aud_bytes = aud_array.tobytes()

        # Construct the payload: header length + header bytes + audio bytes
        payload = header_length_packed + header_bytes + aud_bytes

        # Publish the binary payload
        try:
            MqttMiddlewarePubSub._instance.mqtt_client.publish(self.out_topic, payload)
        except Exception as e:
            logging.error(f"Failed to publish message: {e}")


@Publishers.register("Properties", "mqtt")
class MqttPropertiesPublisher(MqttPublisher):

    def __init__(self, name, out_topic, **kwargs):
        super().__init__(name, out_topic, **kwargs)
        raise NotImplementedError



import argparse

import cv2

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR
from wrapyfi.config.manager import ConfigManager


class ExperimentController(MiddlewareCommunicator):
    WEBCAM_HEIGHT = 240
    WEBCAM_WIDTH = 320
    PEPPER_CAM_HEIGHT = 480
    PEPPER_CAM_WIDTH = 640
    ICUB_CAM_HEIGHT = 240
    ICUB_CAM_WIDTH = 320
    ESR_CAM_HEIGHT = 240
    ESR_CAM_WIDTH = 320

    def __init__(self, **kwargs):
        super(ExperimentController, self).__init__()

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ExperimentController",
        "/control_interface/facial_expressions_esr9",
        should_wait=False,
    )
    def listen_facial_expressions_esr9(self, _mware=DEFAULT_COMMUNICATOR):
        return (None,)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ExperimentController",
        "/control_interface/facial_expressions_icub",
        should_wait=False,
    )
    def publish_facial_expressions_icub(self, obj, _mware="yarp"):
        return (obj,)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$_mware",
        "ExperimentController",
        "/control_interface/facial_expressions_pepper",
        carrier="tcp",
        should_wait=False,
    )
    def publish_facial_expressions_pepper(self, obj, _mware="ros"):
        return (obj,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "$_topic",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
        jpg=True,
    )
    def listen_image_webcam(
        self,
        _mware=DEFAULT_COMMUNICATOR,
        _width=WEBCAM_WIDTH,
        _height=WEBCAM_HEIGHT,
        _topic="/control_interface/image_webcam",
    ):  # dynamic topic according to corresponding API
        return (None,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "$_topic",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
    )
    def listen_image_pepper_cam(
        self,
        _mware="ros",
        _width=PEPPER_CAM_WIDTH,
        _height=PEPPER_CAM_HEIGHT,
        _topic="/pepper/camera/front/camera/image_raw",
    ):
        return (None,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "$_topic",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
    )
    def listen_image_icub_cam(
        self,
        _mware="yarp",
        _width=ICUB_CAM_WIDTH,
        _height=ICUB_CAM_HEIGHT,
        _topic="/icub/cam/right",
    ):
        return (None,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "/control_interface/image_webcam",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
    )
    def forward_image_webcam(
        self,
        img,
        _mware=DEFAULT_COMMUNICATOR,
        _width=WEBCAM_WIDTH,
        _height=WEBCAM_HEIGHT,
    ):
        return (img,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "/control_interface/image_pepper_cam",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        jpg=True,
        should_wait=False,
    )
    def forward_image_pepper_cam(
        self,
        img,
        _mware=DEFAULT_COMMUNICATOR,
        _width=PEPPER_CAM_WIDTH,
        _height=PEPPER_CAM_HEIGHT,
    ):
        return (img,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "/control_interface/image_icub_cam",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        jpg=True,
        should_wait=False,
    )
    def forward_image_icub_cam(
        self,
        img,
        _mware=DEFAULT_COMMUNICATOR,
        _width=ICUB_CAM_WIDTH,
        _height=ICUB_CAM_HEIGHT,
    ):
        # convert to bgr
        if img is not None:
            img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        return (img,)

    @MiddlewareCommunicator.register(
        "Image",
        "$_mware",
        "ExperimentController",
        "/control_interface/image_esr9",
        width="$_width",
        height="$_height",
        rgb=True,
        fp=False,
        should_wait=False,
        jpg=True,
    )
    def publish_esr9_cam(
        self,
        img,
        _mware=DEFAULT_COMMUNICATOR,
        _width=ESR_CAM_WIDTH,
        _height=ESR_CAM_HEIGHT,
    ):
        # modify size of image
        if img is not None:
            img = cv2.resize(img, (self.ESR_CAM_WIDTH, self.ESR_CAM_HEIGHT))
        return (img,)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--wrapyfi_cfg",
        help="File to load Wrapyfi configs for running instance. "
        "Choose one of the configs available in "
        "./wrapyfi_configs/affective_signaling_multirobot "
        "for each running instance. All configs with a prefix of COMP must "
        "have corresponding instances running. OPT prefixed configs execute "
        "scripts on a machine connected to either robot (Pepper/iCub) or both, "
        "and at least one must run in addition to COMP.",
        type=str,
    )
    parser.add_argument(
        "--cam_source",
        help="The camera input source being either from a "
        "webcam, Pepper, or iCub. Note that this must be similar for all running "
        "instances even when the two robots are set to display the expressions.",
        type=str,
        default="webcam",
        choices=["webcam", "pepper", "icub"],
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    if args.wrapyfi_cfg:
        print("Wrapyfi config loading from ", args.wrapyfi_cfg)
        ConfigManager(args.wrapyfi_cfg)

    ec = ExperimentController()
    image_cam = None
    while True:
        if args.cam_source == "webcam":
            (image_webcam,) = ec.listen_image_webcam()
            (image_webcam_linked,) = ec.forward_image_webcam(image_webcam)
            if image_webcam is None:
                image_webcam = image_webcam_linked
            image_cam = image_webcam
            if image_webcam is not None:
                cv2.imshow("Webcam image", image_webcam)
                cv2.waitKey(1)
        if args.cam_source == "pepper":
            (image_pepper_cam,) = ec.listen_image_pepper_cam()
            (image_pepper_cam_linked,) = ec.forward_image_pepper_cam(image_pepper_cam)
            if image_pepper_cam is None:
                image_pepper_cam = image_pepper_cam_linked
            image_cam = image_pepper_cam
            if image_pepper_cam is not None:
                cv2.imshow("Pepper image", image_pepper_cam)
                cv2.waitKey(1)
        if args.cam_source == "icub":
            (image_icub_cam,) = ec.listen_image_icub_cam()
            (image_icub_cam_linked,) = ec.forward_image_icub_cam(image_icub_cam)
            if image_icub_cam is None:
                image_icub_cam = image_icub_cam_linked
            image_cam = image_icub_cam
            if image_icub_cam_linked is not None:
                cv2.imshow("iCub image", image_icub_cam_linked)
                cv2.waitKey(1)

        (image_esr,) = ec.publish_esr9_cam(image_cam)
        (facial_expression,) = ec.listen_facial_expressions_esr9()
        if facial_expression is not None:
            ec.publish_facial_expressions_icub(facial_expression)
            ec.publish_facial_expressions_pepper(facial_expression)



from flask import Flask
from flask_socketio import SocketIO, Namespace
import logging

# Default configuration (should the values not be changed in the environment)
SOCKET_IP = "0.0.0.0"
SOCKET_PORT = 5000
WEBSOCKET_NAMESPACE = "/"

# Initialize Flask and Flask-SocketIO
app = Flask(__name__)
app.config["SECRET_KEY"] = "some secret key"
socketio = SocketIO(app, logger=True, engineio_logger=True)

# Setup logging
logging.basicConfig(level=logging.DEBUG)


# Create a custom namespace class
class MyCustomNamespace(Namespace):
    """
    This is a custom namespace class that extends the Flask-SocketIO Namespace class.
    This class is used to handle custom events and manage connected clients.
    """

    connected_sids = set()

    def on_connect(self, sid, environ, auth):
        logging.debug(f"Client connected: {sid}")
        self.connected_sids.add(sid)
        logging.debug(f"All connected sids: {self.connected_sids}")

    def on_disconnect(self, sid):
        logging.debug(f"Client disconnected: {sid}")
        self.connected_sids.discard(sid)
        logging.debug(f"All connected sids: {self.connected_sids}")

    def trigger_event(self, event, *args):
        logging.debug(f"Received event '{event}' with args: {args}")
        handler = getattr(self, "on_" + event, None)
        if handler is not None:
            return handler(*args)
        else:
            logging.warning(f"No handler for event '{event}'")


# Register the custom namespace
socketio.on_namespace(MyCustomNamespace(WEBSOCKET_NAMESPACE))


@socketio.event(namespace="/")
def connect():
    """
    This is the default event handler for when a client connects to the server.
    """
    print("Connected to the server")


@socketio.on("/hello/my_message", namespace="/")
def my_text_message(data):
    """
    This is for the `websockets/publisher_client.py` example for when the publisher sends a message over Websockets.
    To receive the message, the `websockets/listener_client.py` script must be running.

    :param data: dict: The message data
    """
    print("Received 'my_message' in client script")
    socketio.emit("/hello/my_message", data, namespace="/")


@socketio.on("/cam_mic/cam_feed", namespace="/")
def my_img_message(data):
    """
    This is for the `sensors/cam_mic.py` example for when the publisher sends an image message over websocket (setting `--mware` argument to 'websocket').

    :param data: dict: The image data
    """
    print("Received 'my_message' in client script")
    socketio.emit("/cam_mic/cam_feed", data, namespace="/")


@socketio.on("/cam_mic/audio_feed", namespace="/")
def my_aud_message(data):
    """
    This is for the `sensors/cam_mic.py` example when the publisher sends an audio message over websocket (setting `--mware` argument to 'websocket').

    :param data: dict: The audio data
    """
    print("Received 'my_message' in client script")
    socketio.emit("/cam_mic/audio_feed", data, namespace="/")


# Start the development server. Note that for a production deployment, a production-ready server like waitress should be used.
if __name__ == "__main__":
    socketio.run(
        app, host=SOCKET_IP, port=SOCKET_PORT, debug=True, allow_unsafe_werkzeug=True
    )



import logging
import atexit
import threading

import paho.mqtt.client as mqtt_client

from wrapyfi.utils import SingletonOptimized
from wrapyfi.connect.wrapper import MiddlewareCommunicator


class MqttMiddlewarePubSub(metaclass=SingletonOptimized):
    """
    MQTT middleware wrapper. This class is a singleton, so it can be instantiated only once.
    The `activate` method should be called to initialize the middleware.
    The `deinit` method should be called to deinitialize the middleware and destroy all connections.
    """

    _instance = None  # Singleton instance

    @staticmethod
    def activate(
        broker_address: str = "broker.emqx.io",
        broker_port: int = 1883,
        **kwargs,
    ):
        """
        Activate the MQTT middleware. Initializes the MQTT client with the provided address and options.

        :param broker_address: str: The MQTT broker address
        :param broker_port: int: The MQTT broker port
        :param kwargs: dict: Additional keyword arguments for the middleware
        """
        if MqttMiddlewarePubSub._instance is None:
            MqttMiddlewarePubSub._instance = MqttMiddlewarePubSub(
                broker_address=broker_address, broker_port=broker_port, **kwargs
            )
        return MqttMiddlewarePubSub._instance

    def __init__(
        self,
        broker_address: str = "broker.emqx.io",
        broker_port: int = 1883,
        client_id: str = None,
        **kwargs,
    ):
        """
        Initialize the MQTT middleware. This method is automatically called when the class is instantiated.

        :param broker_address: str: The MQTT broker address
        :param broker_port: int: The MQTT broker port
        :param client_id: str: The MQTT client ID
        :param kwargs: dict: Additional keyword arguments for compatibility with the interface
        """
        logging.info(f"Initializing MQTT middleware on {broker_address}:{broker_port}")

        # Store arguments, even if unused (for compatibility)
        self.broker_address = broker_address
        self.port = broker_port
        self.client_id = client_id

        # Create a MQTT client
        self.mqtt_client = mqtt_client.Client(
            mqtt_client.CallbackAPIVersion.VERSION2, client_id=self.client_id
        )

        # Set up callbacks
        self.mqtt_client.on_connect = self._on_connect
        self.mqtt_client.on_disconnect = self._on_disconnect
        self.mqtt_client.on_message = self._on_message

        # Dictionary to store topic-specific callbacks
        self.topic_callbacks = {}

        # Track connection status
        self.connected = False

        # Start the connection in a background thread
        self.client_thread = threading.Thread(target=self._connect_client)
        self.client_thread.daemon = True
        self.client_thread.start()

        # Ensure cleanup at exit
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

    def _connect_client(self):
        """Connect the MQTT client."""
        try:
            self.mqtt_client.connect(self.broker_address, self.port)
            # Use loop_start instead of loop_forever for non-blocking behavior
            self.mqtt_client.loop_start()
        except Exception as e:
            logging.error(
                f"[MqttMiddlewarePubSub] Error connecting to {self.broker_address}:{self.port}: {e}"
            )

    def _on_connect(self, client, userdata, flags, rc, properties):
        """Callback for when the MQTT client connects to the broker."""
        logging.info(
            f"[MqttMiddlewarePubSub] Connected to {self.broker_address}:{self.port}"
        )
        self.connected = True

    def _on_disconnect(self, client, userdata, rc):
        """Callback for when the MQTT client disconnects from the broker."""
        logging.info(
            f"[MqttMiddlewarePubSub] Disconnected from {self.broker_address}:{self.port}"
        )
        self.connected = False

    def _on_message(self, client, userdata, msg):
        """Callback for when the MQTT client receives a message."""
        if msg.topic in self.topic_callbacks:
            callback = self.topic_callbacks[msg.topic]
            callback(client, userdata, msg)

    def register_callback(self, topic: str, callback):
        """
        Register an event handler for a specific topic.

        :param topic: str: The topic to subscribe to
        :param callback: callable: The function to call when a message is received on this topic
        """
        self.mqtt_client.subscribe(topic)
        self.topic_callbacks[topic] = callback
        logging.info(f"[MqttMiddlewarePubSub] Registered callback for topic {topic}")

    def is_connected(self) -> bool:
        """
        Check whether the MQTT client is connected.

        :return: bool: True if connected, False otherwise.
        """
        return self.connected

    @staticmethod
    def deinit():
        """
        Deinitialize the MQTT middleware. This method is automatically called when the program exits.
        """
        logging.info("[MqttMiddlewarePubSub] Deinitializing MQTT client")
        if MqttMiddlewarePubSub._instance:
            MqttMiddlewarePubSub._instance.mqtt_client.loop_stop()
            MqttMiddlewarePubSub._instance.mqtt_client.disconnect()
            MqttMiddlewarePubSub._instance.connected = False



"""
A message publisher and listener for native Python objects and Dask Arrays/Dataframes.

This script demonstrates the capability to transmit native Python objects and Dask arrays/dataframes using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the PUB/SUB pattern
allowing message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a nested dummy Python object with native objects and Dask arrays/dataframes
    - Applying the PUB/SUB pattern with mirroring

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - Dask, pandas: Used for handling and creating arrays and dataframes

    Install using pip:
        ``pip install pandas dask[complete]``

Run:
    # On machine 1 (or process 1): Publisher waits for keyboard input and transmits message
        ``python3 dask_example.py --mode publish``

    # On machine 2 (or process 2): Listener waits for message and computes and prints the Dask objects
        ``python3 dask_example.py --mode listen``
"""

import argparse

try:
    import dask.array as da
    import dask.dataframe as dd
    import pandas as pd
except ImportError:
    print("Install DASK and pandas before running this script.")

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class Notifier(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "Notifier",
        "/notify/test_dask_exchange",
        carrier="tcp",
        should_wait=True,
    )
    def exchange_object(self, mware=None):
        """
        Exchange messages with Dask arrays/dataframes and other native Python objects.
        """
        msg = input("Type your message: ")

        # Creating an example Dask DataFrame
        df = pd.DataFrame(
            {
                "num_legs": [4, 2, 0, 4],
                "num_wings": [0, 2, 0, 0],
                "num_specimen_seen": [10, 2, 1, 8],
            },
            index=["falcon", "parrot", "fish", "dog"],
        )

        ddf = dd.from_pandas(df, npartitions=2)

        ds = pd.Series([1, 2, 3, 4, 5, 6, 7])
        dds = dd.from_pandas(ds, npartitions=1)
        # Creating an example Dask Array
        darray = da.random.random((1000, 1000), chunks=(250, 250))

        ret = {
            "message": msg,
            "dask_dataframe": ddf,
            "dask_array": darray,
            "dask_series": dds,
        }
        return (ret,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message publisher and listener for native Python objects and Dask arrays/dataframes."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="publish",
        choices={"publish", "listen"},
        help="The transmission mode",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate Notifier class and communication.
    """
    notifier = Notifier()
    notifier.activate_communication(Notifier.exchange_object, mode=args.mode)

    while True:
        (msg_object,) = notifier.exchange_object(mware=args.mware)

        # Compute and print the actual values of the Dask objects
        for key, value in msg_object.items():
            if isinstance(value, (dd.DataFrame, dd.Series, da.Array)):
                print(f"{key} computed: \n{value.compute()}")
            else:
                print(f"{key}: {value}")


if __name__ == "__main__":
    args = parse_args()
    main(args)



import logging
import atexit
import threading

import socketio

from wrapyfi.utils import SingletonOptimized
from wrapyfi.connect.wrapper import MiddlewareCommunicator


class WebSocketMiddlewarePubSub(metaclass=SingletonOptimized):
    """
    WebSocket middleware wrapper. This class is a singleton, so it can be instantiated only once.
    The `activate` method should be called to initialize the middleware.
    The `deinit` method should be called to deinitialize the middleware and destroy all connections.
    """

    _instance = None  # Singleton instance

    @staticmethod
    def activate(socket_address: str = "http://127.0.0.1:5000", **kwargs):
        """
        Activate the WebSocket middleware. Initializes the WebSocket client with the provided address and options.

        :param socket_address: str: The WebSocket server address
        :param kwargs: dict: Additional keyword arguments for the middleware
        """
        if WebSocketMiddlewarePubSub._instance is None:
            WebSocketMiddlewarePubSub._instance = WebSocketMiddlewarePubSub(
                socket_address=socket_address, **kwargs
            )
        return WebSocketMiddlewarePubSub._instance

    def __init__(
        self,
        socket_address: str = "http://127.0.0.1:5000",
        monitor_listener_spawn: str = None,
        websocket_kwargs: dict = None,
        **kwargs,
    ):
        """
        Initialize the WebSocket middleware. This method is automatically called when the class is instantiated.

        :param socket_address: str: The WebSocket server address
        :param monitor_listener_spawn: str: Determines the type of listener spawn
        :param websocket_kwargs: dict: Additional keyword arguments for the WebSocket connection
        :param kwargs: dict: Additional keyword arguments for compatibility with the interface
        """
        logging.info(f"Initializing WebSocket middleware on {socket_address}")

        # Store arguments, even if unused for now (for interface compatibility)
        self.socket_address = socket_address
        self.monitor_listener_spawn = monitor_listener_spawn
        self.websocket_kwargs = websocket_kwargs or {}

        # Initialize WebSocket client
        self.socketio_client = socketio.Client()

        # Track connection status
        self.connected = False

        # Register event handlers for connection
        @self.socketio_client.event
        def connect():
            logging.info(
                f"[WebSocketMiddlewarePubSub] Connected to {self.socket_address}"
            )
            self.connected = True

        @self.socketio_client.event
        def disconnect():
            logging.info(
                f"[WebSocketMiddlewarePubSub] Disconnected from {self.socket_address}"
            )
            self.connected = False

        # Start the connection in a background thread
        self.client_thread = threading.Thread(target=self._connect_client)
        self.client_thread.daemon = True
        self.client_thread.start()

        # Ensure cleanup at exit
        atexit.register(MiddlewareCommunicator.close_all_instances)
        atexit.register(self.deinit)

    def _connect_client(self):
        """Connect the WebSocket client."""
        try:
            self.socketio_client.connect(
                self.socket_address, namespaces=["/"], **self.websocket_kwargs
            )
            self.socketio_client.wait()  # Wait for messages
        except Exception as e:
            logging.error(
                f"[WebSocketMiddlewarePubSub] Error connecting to {self.socket_address}: {e}"
            )

    def register_callback(self, topic: str, callback):
        """
        Register an event handler for a specific topic.

        :param topic: str: The topic/event to listen to
        :param callback: callable: The function to call when the event occurs
        """
        self.socketio_client.on(topic, callback)
        logging.info(
            f"[WebSocketMiddlewarePubSub] Registered callback for topic {topic}"
        )

    def is_connected(self) -> bool:
        """
        Check whether the WebSocket client is connected.

        :return: bool: True if connected, False otherwise.
        """
        return self.connected

    @staticmethod
    def deinit():
        """
        Deinitialize the WebSocket middleware. This method is automatically called when the program exits.
        """
        logging.info("[WebSocketMiddlewarePubSub] Deinitializing WebSocket client")
        if (
            WebSocketMiddlewarePubSub._instance
            and WebSocketMiddlewarePubSub._instance.socketio_client.connected
        ):
            WebSocketMiddlewarePubSub._instance.socketio_client.disconnect()
            WebSocketMiddlewarePubSub._instance.connected = False



import logging
import time
import os
import queue
import json
from typing import Optional, Tuple
import random
import struct

import numpy as np
import cv2

from wrapyfi.connect.listeners import Listener, Listeners, ListenerWatchDog
from wrapyfi.encoders import JsonDecodeHook
from wrapyfi.middlewares.mqtt import MqttMiddlewarePubSub

MQTT_BROKER_ADDRESS = os.environ.get("WRAPYFI_MQTT_BROKER_ADDRESS", "broker.emqx.io")
MQTT_BROKER_PORT = int(os.environ.get("WRAPYFI_MQTT_BROKER_PORT", 1883))
WATCHDOG_POLL_REPEAT = None


class MqttListener(Listener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        broker_address: str = MQTT_BROKER_ADDRESS,
        broker_port: int = MQTT_BROKER_PORT,
        mqtt_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the subscriber.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param broker_address: str: Address of the MQTT broker. Default is 'broker.emqx.io'
        :param broker_port: int: Port of the MQTT broker. Default is 1883
        :param mqtt_kwargs: dict: Additional kwargs for the MQTT middleware
        :param kwargs: dict: Additional kwargs for the subscriber
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)

        if mqtt_kwargs is None:
            mqtt_kwargs = {}

        if "client_id" not in mqtt_kwargs:
            mqtt_kwargs["client_id"] = f"client-{name}-{random.randint(0, 100000)}"
        mqtt_kwargs["clean_session"] = False

        # Activate the MQTT middleware
        MqttMiddlewarePubSub.activate(
            broker_address=broker_address,
            broker_port=broker_port,
            **(mqtt_kwargs or {}),
        )

        if not self.should_wait:
            ListenerWatchDog().add_listener(self)

    def await_connection(
        self, in_topic: Optional[str] = None, repeats: Optional[int] = None
    ):
        """
        Wait until the MQTT connection is established.

        :param in_topic: str: The topic to monitor for connection
        :param repeats: int: The number of times to check for the connection, None for infinite.
        """
        if in_topic is None:
            in_topic = self.in_topic
        logging.info(f"[MQTT] Waiting for input topic: {in_topic}")
        if repeats is None:
            repeats = -1 if self.should_wait else 0

        # Ensure to call is_connected() on the singleton instance
        while repeats > 0 or repeats == -1:
            if repeats != -1:
                repeats -= 1
            connected = (
                MqttMiddlewarePubSub._instance.is_connected()
            )  # Use the instance
            logging.debug(f"Connection status: {connected}")
            if connected:
                logging.info(f"[MQTT] Connected to input topic: {in_topic}")
                return True
            time.sleep(0.2)
        return False

    def close(self):
        """
        Close the subscriber.
        """
        pass

    def __del__(self):
        self.close()


@Listeners.register("NativeObject", "mqtt")
class MqttNativeObjectListener(MqttListener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        deserializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObject listener using the MQTT message construct assuming the data is serialized as a JSON string.
        Deserializes the data (including plugins) using the decoder and parses it to a native object.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param deserializer_kwargs: dict: Additional kwargs for the deserializer
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)
        self._plugin_decoder_hook = JsonDecodeHook(**kwargs).object_hook
        self._deserializer_kwargs = deserializer_kwargs or {}
        self._message_queue = queue.Queue()

    def on_message(self, client, userdata, msg):
        """
        Callback for handling incoming MQTT messages.
        """
        try:
            obj = json.loads(
                msg.payload.decode(),
                object_hook=self._plugin_decoder_hook,
                **self._deserializer_kwargs,
            )
            self._message_queue.put(obj)
            logging.debug(f"Message queued for topic {self.in_topic}: {obj}")
        except json.JSONDecodeError as e:
            logging.error(f"Failed to decode JSON from topic {self.in_topic}: {e}")

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(repeats=repeats)
        established = self.check_establishment(established)
        if established:
            MqttMiddlewarePubSub._instance.register_callback(
                self.in_topic, self.on_message
            )
        return established

    def listen(self):
        """
        Listen for a message.

        :return: Any: The received message as a native Python object
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None

        try:
            obj = self._message_queue.get(block=self.should_wait)
            return obj
        except queue.Empty:
            return None


@Listeners.register("Image", "mqtt")
class MqttImageListener(MqttNativeObjectListener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        **kwargs,
    ):
        """
        The Image listener using the MQTT message construct parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param width: int: Width of the image. Default is -1 (use the width of the received image)
        :param height: int: Height of the image. Default is -1 (use the height of the received image)
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be decompressed from JPG. Default is False
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8
        self._message_queue = queue.Queue()

    def on_message(self, client, userdata, msg):
        """
        Callback for handling incoming image messages.
        """
        try:
            payload = msg.payload
            # Read the first 4 bytes to get the header length
            header_length_packed = payload[:4]
            header_length = struct.unpack("!I", header_length_packed)[0]
            # Read the header bytes
            header_bytes = payload[4 : 4 + header_length]
            header_json = header_bytes.decode("utf-8")
            header = json.loads(header_json)
            # Remaining bytes are image bytes
            img_bytes = payload[4 + header_length :]

            if self.jpg:
                # JPEG case: decode the JPEG image
                img_array = np.frombuffer(img_bytes, dtype=np.uint8)
                if self.rgb:
                    img = cv2.imdecode(img_array, cv2.IMREAD_COLOR)
                else:
                    img = cv2.imdecode(img_array, cv2.IMREAD_GRAYSCALE)
                self._message_queue.put(img)
            else:
                # Non-JPEG case: reconstruct the numpy array
                shape = header.get("shape", None)
                dtype = header.get("dtype", None)
                if shape is not None and dtype is not None:
                    img_array = np.frombuffer(img_bytes, dtype=dtype)
                    img = img_array.reshape(shape)
                    self._message_queue.put(img)
                else:
                    logging.error(
                        "Missing 'shape' or 'dtype' in header for non-JPEG image"
                    )
        except Exception as e:
            logging.error(f"Failed to process message from topic {self.in_topic}: {e}")

    def listen(self) -> Optional[np.ndarray]:
        """
        Listen for a message.

        :return: np.ndarray: The received image as a numpy array formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None

        try:
            img = self._message_queue.get(block=self.should_wait)
            if (
                (self.width > 0 and self.width != img.shape[1])
                or (self.height > 0 and self.height != img.shape[0])
                or not (
                    (img.ndim == 2 and not self.rgb)
                    or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
                )
            ):
                raise ValueError("Incorrect image shape for listener")
            return img
        except queue.Empty:
            return None


@Listeners.register("AudioChunk", "mqtt")
class MqttAudioChunkListener(MqttNativeObjectListener):

    def __init__(
        self,
        name: str,
        in_topic: str,
        should_wait: bool = True,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        **kwargs,
    ):
        """
        The AudioChunk listener using the MQTT message construct parsed to a numpy array.

        :param name: str: Name of the subscriber
        :param in_topic: str: Name of the input topic
        :param should_wait: bool: Whether the subscriber should wait for the publisher to transmit a message. Default is True
        :param channels: int: Number of channels in the audio. Default is 1
        :param rate: int: Sampling rate of the audio. Default is 44100
        :param chunk: int: Number of samples in the audio chunk. Default is -1 (use the chunk size of the received audio)
        """
        super().__init__(name, in_topic, should_wait=should_wait, **kwargs)
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

        self._message_queue = queue.Queue()

    def on_message(self, client, userdata, msg):
        """
        Callback for handling incoming audio chunk messages.
        """
        try:
            payload = msg.payload
            # Ensure payload is long enough to contain the header length
            if len(payload) < 4:
                logging.error("Payload too short to contain header length")
                return

            # Read the first 4 bytes to get the header length
            header_length_packed = payload[:4]
            header_length = struct.unpack("!I", header_length_packed)[0]

            # Ensure payload is long enough to contain the header
            if len(payload) < 4 + header_length:
                logging.error("Payload too short to contain header")
                return

            # Read the header bytes
            header_bytes = payload[4 : 4 + header_length]
            header_json = header_bytes.decode("utf-8")
            header = json.loads(header_json)

            # Remaining bytes are audio bytes
            aud_bytes = payload[4 + header_length :]

            # Extract metadata from header
            shape = header.get("shape")
            dtype = header.get("dtype")
            rate = header.get("rate")
            timestamp = header.get("timestamp")

            if shape is None or dtype is None or rate is None:
                logging.error("Missing 'shape', 'dtype', or 'rate' in header")
                return

            if 0 < self.rate != rate:
                raise ValueError("Incorrect audio rate for listener")

            # Reconstruct the audio array from the binary data
            aud_array = np.frombuffer(aud_bytes, dtype=dtype)
            if aud_array.size != np.prod(shape):
                logging.error("Mismatch between audio data size and shape")
                return

            aud_array = aud_array.reshape(shape)

            chunk, channels = (
                aud_array.shape if len(aud_array.shape) > 1 else (aud_array.shape[0], 1)
            )

            if (0 < self.chunk != chunk) or (0 < self.channels != channels):
                raise ValueError("Incorrect audio shape for listener")

            self._message_queue.put((aud_array, rate))
        except Exception as e:
            logging.error(f"Failed to process message from topic {self.in_topic}: {e}")

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        return super().establish(repeats=repeats, **kwargs)

    def listen(self) -> Tuple[Optional[np.ndarray], int]:
        """
        Listen for a message.

        :return: Tuple[np.ndarray, int]: The received audio chunk as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return None, self.rate

        try:
            aud, rate = self._message_queue.get(block=self.should_wait)
            return aud, rate
        except queue.Empty:
            return None, self.rate


@Listeners.register("Properties", "mqtt")
class MqttPropertiesListener(MqttNativeObjectListener):
    def __init__(self, name, in_topic, **kwargs):
        super().__init__(name, in_topic, **kwargs)
        raise NotImplementedError



import os
from functools import wraps
import re
from typing import Union, List, Any, Callable, Optional

import wrapyfi.connect.publishers as pub
import wrapyfi.connect.listeners as lsn
import wrapyfi.connect.servers as srv
import wrapyfi.connect.clients as clt

from wrapyfi.utils import get_default_args, match_args, deepcopy
from wrapyfi.config.manager import ConfigManager


DEFAULT_COMMUNICATOR = os.environ.get("WRAPYFI_DEFAULT_COMMUNICATOR", "zeromq")
DEFAULT_COMMUNICATOR = os.environ.get("WRAPYFI_DEFAULT_MWARE", DEFAULT_COMMUNICATOR)


class MiddlewareCommunicator(object):
    __registry = {}

    def __init__(self):
        """
        Initialises the middleware communicator.
        """
        self.config = ConfigManager(None).config
        if self.__class__.__name__ in self.config:
            for key, value in self.config[self.__class__.__name__].items():
                self.activate_communication(getattr(self.__class__, key), mode=value)

    @classmethod
    def __trigger_publish(
        cls, func: Callable[..., Any], instance_id: str, kwd: dict, *wds, **kwds
    ):
        """
        Triggers the publish mode of the middleware communicator. If the intended publisher type and middleware are unavailable, resorts to a fallback publisher instead.

        :param func: Callable[..., Any]: The function to be triggered and whose return values are to be published
        :param instance_id: str: The unique identifier of the function instance, utilized to access the middleware communicator registry and manage different instances of communications
        :param kwd: dict: A dictionary containing keyword arguments to be matched and potentially utilized during the publisher instantiation and function invocation
        :param wds: tuple: Variable positional arguments to be passed to the function
        :param kwds: dict: Additional keyword arguments to be passed to the function
        :return: Any: The return values of the triggered function (`func`). The data type and structure depend on the output of `func`
        """
        if (
            "wrapped_executor"
            not in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"][0]
        ):
            # instantiate the publishers
            cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                "communicator"
            ].reverse()
            for communicator in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"]:
                # single element
                if isinstance(communicator["return_func_type"], str):
                    return_func_pub_kwargs = deepcopy(
                        communicator["return_func_kwargs"]
                    )
                    return_func_pub_kwargs.update(
                        return_func_pub_kwargs.get("publisher_kwargs", {})
                    )
                    return_func_pub_kwargs.pop("listener_kwargs", None)
                    return_func_pub_kwargs.pop("publisher_kwargs", None)
                    new_args, new_kwargs = match_args(
                        communicator["return_func_args"],
                        return_func_pub_kwargs,
                        wds[1:],
                        kwd,
                    )
                    return_func_type = communicator["return_func_type"]
                    return_func_middleware = new_kwargs.pop(
                        "middleware", DEFAULT_COMMUNICATOR
                    )
                    try:
                        communicator["wrapped_executor"] = pub.Publishers.registry[
                            return_func_type + return_func_middleware
                        ](*new_args, **new_kwargs)
                    except KeyError:
                        communicator["wrapped_executor"] = pub.Publishers.registry[
                            "MMO:fallback"
                        ](
                            *new_args,
                            missing_middleware_object=return_func_type
                            + return_func_middleware,
                            **new_kwargs,
                        )
                        communicator["return_func_type"] = "MMO:"
                # list for single return
                elif isinstance(communicator["return_func_type"], list):
                    communicator["wrapped_executor"] = []
                    for comm_idx in range(len(communicator["return_func_type"])):
                        return_func_pub_kwargs = deepcopy(
                            communicator["return_func_kwargs"][comm_idx]
                        )
                        return_func_pub_kwargs.update(
                            return_func_pub_kwargs.get("publisher_kwargs", {})
                        )
                        return_func_pub_kwargs.pop("listener_kwargs", None)
                        return_func_pub_kwargs.pop("publisher_kwargs", None)
                        new_args, new_kwargs = match_args(
                            communicator["return_func_args"][comm_idx],
                            return_func_pub_kwargs,
                            wds[1:],
                            kwd,
                        )
                        return_func_type = communicator["return_func_type"][comm_idx]
                        return_func_middleware = new_kwargs.pop(
                            "middleware", DEFAULT_COMMUNICATOR
                        )
                        try:
                            communicator["wrapped_executor"].append(
                                pub.Publishers.registry[
                                    return_func_type + return_func_middleware
                                ](*new_args, **new_kwargs)
                            )
                        except KeyError:
                            communicator["wrapped_executor"].append(
                                pub.Publishers.registry["MMO:fallback"](
                                    *new_args,
                                    missing_middleware_object=return_func_type
                                    + return_func_middleware,
                                    **new_kwargs,
                                )
                            )
                            communicator["return_func_type"][comm_idx] = "MMO:"

        returns = func(*wds, **kwds)
        for ret_idx, ret in enumerate(returns):
            wrp_exec = cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"][ret_idx]["wrapped_executor"]
            # single element
            if isinstance(wrp_exec, pub.Publisher):
                wrp_exec.publish(ret)
            # list for single return
            elif isinstance(wrp_exec, list):
                for wrp_idx, wrp in enumerate(wrp_exec):
                    wrp.publish(ret[wrp_idx])
        return returns

    @classmethod
    def __trigger_listen(
        cls, func: Callable[..., Any], instance_id: str, kwd: dict, *wds, **kwds
    ):
        """
        Triggers the listen mode of the middleware communicator. If the intended listener type and middleware are unavailable, resorts to a fallback listener instead.

        :param func: Callable[..., Any]: The function associated with the listener and whose return values might be utilized
        :param instance_id: str: The unique identifier of the function instance, utilized to access the middleware communicator registry and manage different instances of communications
        :param kwd: dict: A dictionary containing keyword arguments to be matched and potentially utilized during the listener instantiation and function invocation
        :param wds: tuple: Variable positional arguments to be passed to the function
        :param kwds: dict: Additional keyword arguments to be passed to the function (not used since the listener does not accept any additional arguments, given that it does not execute any function or transmit any arguments)
        :return: Any: The return values obtained from the listeners. The data type and structure depend on the output of the listener
        """
        if (
            "wrapped_executor"
            not in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"][0]
        ):
            # instantiate the listeners
            cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                "communicator"
            ].reverse()
            for communicator in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"]:
                # single element
                if isinstance(communicator["return_func_type"], str):
                    return_func_lsn_kwargs = deepcopy(
                        communicator["return_func_kwargs"]
                    )
                    return_func_lsn_kwargs.update(
                        return_func_lsn_kwargs.get("listener_kwargs", {})
                    )
                    return_func_lsn_kwargs.pop("listener_kwargs", None)
                    return_func_lsn_kwargs.pop("publisher_kwargs", None)
                    new_args, new_kwargs = match_args(
                        communicator["return_func_args"],
                        return_func_lsn_kwargs,
                        wds[1:],
                        kwd,
                    )
                    return_func_type = communicator["return_func_type"]
                    return_func_middleware = new_kwargs.pop(
                        "middleware", DEFAULT_COMMUNICATOR
                    )
                    try:
                        communicator["wrapped_executor"] = lsn.Listeners.registry[
                            return_func_type + return_func_middleware
                        ](*new_args, **new_kwargs)
                    except KeyError:
                        communicator["wrapped_executor"] = lsn.Listeners.registry[
                            "MMO:fallback"
                        ](
                            *new_args,
                            missing_middleware_object=return_func_type
                            + return_func_middleware,
                            **new_kwargs,
                        )
                        communicator["return_func_type"] = "MMO:"

                # list for single return
                elif isinstance(communicator["return_func_type"], list):
                    communicator["wrapped_executor"] = []
                    for comm_idx in range(len(communicator["return_func_type"])):
                        return_func_lsn_kwargs = deepcopy(
                            communicator["return_func_kwargs"][comm_idx]
                        )
                        return_func_lsn_kwargs.update(
                            return_func_lsn_kwargs.get("listener_kwargs", {})
                        )
                        return_func_lsn_kwargs.pop("listener_kwargs", None)
                        return_func_lsn_kwargs.pop("publisher_kwargs", None)
                        new_args, new_kwargs = match_args(
                            communicator["return_func_args"][comm_idx],
                            return_func_lsn_kwargs,
                            wds[1:],
                            kwd,
                        )
                        return_func_type = communicator["return_func_type"][comm_idx]
                        return_func_middleware = new_kwargs.pop(
                            "middleware", DEFAULT_COMMUNICATOR
                        )
                        try:
                            communicator["wrapped_executor"].append(
                                lsn.Listeners.registry[
                                    return_func_type + return_func_middleware
                                ](*new_args, **new_kwargs)
                            )
                        except KeyError:
                            communicator["wrapped_executor"].append(
                                lsn.Listeners.registry["MMO:fallback"](
                                    *new_args,
                                    missing_middleware_object=return_func_type
                                    + return_func_middleware,
                                    **new_kwargs,
                                )
                            )
                            communicator["return_func_type"][comm_idx] = "MMO:"

        returns = []
        for ret_idx in range(
            len(
                cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                    "communicator"
                ]
            )
        ):
            wrp_exec = cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"][ret_idx]["wrapped_executor"]
            # single element
            if isinstance(wrp_exec, lsn.Listener):
                returns.append(wrp_exec.listen())
                # list for single return
            elif isinstance(wrp_exec, list):
                subreturns = []
                for wrp_idx, wrp in enumerate(wrp_exec):
                    subreturns.append(wrp.listen())
                returns.append(subreturns)
        return returns

    @classmethod
    def __trigger_reply(
        cls, func: Callable[..., Any], instance_id: str, kwd, *wds, **kwds
    ):
        """
        Triggers the reply mode of the middleware communicator. If the intended server type and middleware are unavailable, resorts to a fallback server instead.

        :param func: Callable[..., Any]: The function to be triggered and whose return values may be used in replies
        :param instance_id: str: The unique identifier of the function instance, utilized to access the middleware communicator registry and manage different instances of communications
        :param kwd: dict: A dictionary containing keyword arguments to be matched and potentially utilized during the server instantiation and function invocation
        :param wds: tuple: Variable positional arguments to be passed to the function
        :param kwds: dict: Additional keyword arguments to be passed to the function
        :return: Any: The return values of the triggered function (`func`). The data type and structure depend on the output of `func`
        """
        if (
            "wrapped_executor"
            not in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"][0]
        ):
            # instantiate the publishers
            cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                "communicator"
            ].reverse()
            for communicator in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"]:
                # single element
                if isinstance(communicator["return_func_type"], str):
                    return_func_pub_kwargs = deepcopy(
                        communicator["return_func_kwargs"]
                    )
                    return_func_pub_kwargs.update(
                        return_func_pub_kwargs.get("publisher_kwargs", {})
                    )
                    return_func_pub_kwargs.pop("listener_kwargs", None)
                    return_func_pub_kwargs.pop("publisher_kwargs", None)
                    new_args, new_kwargs = match_args(
                        communicator["return_func_args"],
                        return_func_pub_kwargs,
                        wds[1:],
                        kwd,
                    )
                    return_func_type = communicator["return_func_type"]
                    return_func_middleware = new_kwargs.pop(
                        "middleware", DEFAULT_COMMUNICATOR
                    )
                    try:
                        communicator["wrapped_executor"] = srv.Servers.registry[
                            return_func_type + return_func_middleware
                        ](*new_args, **new_kwargs)
                    except KeyError:
                        communicator["wrapped_executor"] = srv.Servers.registry[
                            "MMO:fallback"
                        ](
                            *new_args,
                            missing_middleware_object=return_func_type
                            + return_func_middleware,
                            **new_kwargs,
                        )
                        communicator["return_func_type"] = "MMO:"

                # list for single return
                elif isinstance(communicator["return_func_type"], list):
                    communicator["wrapped_executor"] = []
                    for comm_idx in range(len(communicator["return_func_type"])):
                        return_func_pub_kwargs = deepcopy(
                            communicator["return_func_kwargs"][comm_idx]
                        )
                        return_func_pub_kwargs.update(
                            return_func_pub_kwargs.get("publisher_kwargs", {})
                        )
                        return_func_pub_kwargs.pop("listener_kwargs", None)
                        return_func_pub_kwargs.pop("publisher_kwargs", None)
                        new_args, new_kwargs = match_args(
                            communicator["return_func_args"][comm_idx],
                            return_func_pub_kwargs,
                            wds[1:],
                            kwd,
                        )
                        return_func_type = communicator["return_func_type"][comm_idx]
                        return_func_middleware = new_kwargs.pop(
                            "middleware", DEFAULT_COMMUNICATOR
                        )
                        try:
                            communicator["wrapped_executor"].append(
                                srv.Servers.registry[
                                    return_func_type + return_func_middleware
                                ](*new_args, **new_kwargs)
                            )
                        except KeyError:
                            communicator["wrapped_executor"].append(
                                srv.Servers.registry["MMO:fallback"](
                                    *new_args,
                                    missing_middleware_object=return_func_type
                                    + return_func_middleware,
                                    **new_kwargs,
                                )
                            )
                            communicator["return_func_type"][comm_idx] = "MMO:"

        returns = None
        for ret_idx, functor in enumerate(
            cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                "communicator"
            ]
        ):
            wrp_exec = functor["wrapped_executor"]
            # single element
            if isinstance(wrp_exec, srv.Server):
                new_args, new_kwargs = wrp_exec.await_request(*wds[1:], **kwds)
                if returns is None:
                    returns = func(wds[0], *new_args, **new_kwargs)
                ret = returns[ret_idx]
                wrp_exec.reply(ret)
            # list for single return
            elif isinstance(wrp_exec, list):
                for wrp_idx, wrp in enumerate(wrp_exec):
                    new_args, new_kwargs = wrp.await_request(*wds[1:], **kwds)
                    if returns is None:
                        returns = func(wds[0], *new_args, **new_kwargs)
                    ret = returns[ret_idx][wrp_idx]
                    wrp.reply(ret)
        return returns

    @classmethod
    def __trigger_request(
        cls, func: Callable[..., Any], instance_id: str, kwd, *wds, **kwds
    ):
        """
        Triggers the request mode of the middleware communicator. If the intended client type and middleware are unavailable, resorts to a fallback client instead.

        :param func: Callable[..., Any]: The function associated with the client requests and might utilize the request results
        :param instance_id: str: The unique identifier of the function instance, utilized to access the middleware communicator registry and manage different instances of communications
        :param kwd: dict: A dictionary containing keyword arguments to be matched and potentially utilized during the client instantiation and function invocation
        :param wds: tuple: Variable positional arguments to be passed to the function
        :param kwds: dict: Additional keyword arguments to be passed to the function
        :return: Any: The return values obtained from the clients' requests. The data type and structure depend on the output of the client requests
        """
        if (
            "wrapped_executor"
            not in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"][0]
        ):
            # instantiate the listeners
            cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                "communicator"
            ].reverse()
            for communicator in cls._MiddlewareCommunicator__registry[
                func.__qualname__ + instance_id
            ]["communicator"]:
                # single element
                if isinstance(communicator["return_func_type"], str):
                    return_func_lsn_kwargs = deepcopy(
                        communicator["return_func_kwargs"]
                    )
                    return_func_lsn_kwargs.update(
                        return_func_lsn_kwargs.get("listener_kwargs", {})
                    )
                    return_func_lsn_kwargs.pop("listener_kwargs", None)
                    return_func_lsn_kwargs.pop("publisher_kwargs", None)
                    new_args, new_kwargs = match_args(
                        communicator["return_func_args"],
                        return_func_lsn_kwargs,
                        wds[1:],
                        kwd,
                    )
                    return_func_type = communicator["return_func_type"]
                    return_func_middleware = new_kwargs.pop(
                        "middleware", DEFAULT_COMMUNICATOR
                    )
                    try:
                        communicator["wrapped_executor"] = clt.Clients.registry[
                            return_func_type + return_func_middleware
                        ](*new_args, **new_kwargs)
                    except KeyError:
                        communicator["wrapped_executor"] = clt.Clients.registry[
                            "MMO:fallback"
                        ](
                            *new_args,
                            missing_middleware_object=return_func_type
                            + return_func_middleware,
                            **new_kwargs,
                        )
                        communicator["return_func_type"] = "MMO:"

                # list for single return
                elif isinstance(communicator["return_func_type"], list):
                    communicator["wrapped_executor"] = []
                    for comm_idx in range(len(communicator["return_func_type"])):
                        return_func_lsn_kwargs = deepcopy(
                            communicator["return_func_kwargs"][comm_idx]
                        )
                        return_func_lsn_kwargs.update(
                            return_func_lsn_kwargs.get("listener_kwargs", {})
                        )
                        return_func_lsn_kwargs.pop("listener_kwargs", None)
                        return_func_lsn_kwargs.pop("publisher_kwargs", None)
                        new_args, new_kwargs = match_args(
                            communicator["return_func_args"][comm_idx],
                            return_func_lsn_kwargs,
                            wds[1:],
                            kwd,
                        )
                        return_func_type = communicator["return_func_type"][comm_idx]
                        return_func_middleware = new_kwargs.pop(
                            "middleware", DEFAULT_COMMUNICATOR
                        )
                        try:
                            communicator["wrapped_executor"].append(
                                clt.Clients.registry[
                                    return_func_type + return_func_middleware
                                ](*new_args, **new_kwargs)
                            )
                        except KeyError:
                            communicator["wrapped_executor"].append(
                                clt.Clients.registry["MMO:fallback"](
                                    *new_args,
                                    missing_middleware_object=return_func_type
                                    + return_func_middleware,
                                    **new_kwargs,
                                )
                            )
                            communicator["return_func_type"][comm_idx] = "MMO:"

        returns = []
        for ret_idx, functor in enumerate(
            cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                "communicator"
            ]
        ):
            wrp_exec = functor["wrapped_executor"]
            # single element
            if isinstance(wrp_exec, clt.Client):
                ret = wrp_exec.request(*wds[1:], **kwds)
                returns.append(ret)
            # list for single return
            elif isinstance(wrp_exec, list):
                subreturns = []
                for wrp_idx, wrp in enumerate(wrp_exec):
                    ret = wrp.request(*wds[1:], **kwds)
                    subreturns.append(ret[wrp_idx])
                returns.append(subreturns)
        return returns

    @classmethod
    def register(
        cls,
        data_type: Union[str, List[Any]],
        middleware: str = DEFAULT_COMMUNICATOR,
        *args,
        **kwargs,
    ):
        """
        Registers a function to the middleware communicator, defining its communication message type and associated
        middleware. Note that the function returned is a wrapper that can alter the behavior of the registered function
        based on the communication mode set in the middleware communicator registry.

        :param data_type: Union[str, List[Any]]: Specifies the communication message type, either as a string or a list containing specifications of the data type and middleware
        :param middleware: str: Specifies the middleware to be used for the communication, defaults to DEFAULT_COMMUNICATOR
        :param args: tuple: Variable positional arguments to be passed to the function
        :param kwargs: dict: Additional keyword arguments to be passed to the function
        :return: Callable[..., Any]: A wrapper function that triggers specific communication modes (e.g., publish, listen, reply, request) based on the registered function and middleware settings

        :raises: NotImplementedError: If `data_type` is a dictionary or an unsupported type
        """

        def encapsulate(func):
            # define the communication message type (single element)
            if isinstance(data_type, str):
                return_func_type = data_type + ":"
                return_func_args = args
                return_func_kwargs = kwargs
                return_func_kwargs["middleware"] = str(middleware)

            # define the communication message type (list for single return). NOTE: supports 1 layer depth only
            elif isinstance(data_type, list):
                return_func_args, return_func_kwargs, return_func_type = [], [], []
                for arg in data_type:
                    data_spec = arg[0] + ":"
                    return_func_args.append(
                        [a for a in arg[2:] if not isinstance(a, dict)]
                    )
                    return_func_kwargs.append(
                        *[a for a in arg[2:] if isinstance(a, dict)]
                    )
                    return_func_kwargs[-1]["middleware"] = str(arg[1])
                    return_func_type.append(data_spec)

            # define the communication message type (dict for single return). NOTE: supports 1 layer depth only
            elif isinstance(data_type, dict):
                raise NotImplementedError(
                    "Dictionaries are not yet supported as a return type"
                )

            else:
                raise NotImplementedError(
                    f"Return data type not supported: {data_type}"
                )

            func_qualname = func.__qualname__
            if func_qualname in cls.__registry:
                cls.__registry[func_qualname]["communicator"].append(
                    {
                        "return_func_args": return_func_args,
                        "return_func_kwargs": return_func_kwargs,
                        "return_func_type": return_func_type,
                    }
                )
            else:
                cls.__registry[func_qualname] = {
                    "communicator": [
                        {
                            "return_func_args": return_func_args,
                            "return_func_kwargs": return_func_kwargs,
                            "return_func_type": return_func_type,
                        }
                    ]
                }
            cls.__registry[func_qualname]["mode"] = None

            @wraps(func)
            def wrapper(*wds, **kwds):  # triggers on calling the method
                if hasattr(func, "__wrapped__"):
                    return func(*wds, **kwds)

                instance_address = hex(id(wds[0]))
                try:
                    instance_id = (
                        cls._MiddlewareCommunicator__registry[func.__qualname__][
                            "__WRAPYFI_INSTANCES"
                        ].index(instance_address)
                        + 1
                    )
                    instance_id = "" if instance_id <= 1 else "." + str(instance_id)
                except KeyError:
                    instance_id = ""

                # execute the method as usual
                if (
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["mode"]
                    is None
                ):
                    return func(*wds, **kwds)

                kwd = get_default_args(func)
                kwd.update(kwds)
                cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                    "args"
                ] = wds
                cls._MiddlewareCommunicator__registry[func.__qualname__ + instance_id][
                    "kwargs"
                ] = kwd

                # publishes the method returns
                if (
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["mode"]
                    == "publish"
                ):
                    return cls.__trigger_publish(func, instance_id, kwd, *wds, **kwds)

                # listens to the publisher and returns the messages
                elif (
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["mode"]
                    == "listen"
                ):
                    return cls.__trigger_listen(func, instance_id, kwd, *wds, **kwds)

                # server awaits request from client and replies with method returns
                elif (
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["mode"]
                    == "reply"
                ):
                    return cls.__trigger_reply(func, instance_id, kwd, *wds, **kwds)

                # client requests with args from server and awaits reply
                elif (
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["mode"]
                    == "request"
                ):
                    return cls.__trigger_request(func, instance_id, kwd, *wds, **kwds)

                # WARNING: use with caution. This produces "None" for all the method's returns
                elif (
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["mode"]
                    == "disable"
                ):
                    cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["last_results"] = []
                    for ret_idx in range(
                        len(
                            cls._MiddlewareCommunicator__registry[
                                func.__qualname__ + instance_id
                            ]["communicator"]
                        )
                    ):
                        cls._MiddlewareCommunicator__registry[
                            func.__qualname__ + instance_id
                        ]["last_results"].append(None)
                    return cls._MiddlewareCommunicator__registry[
                        func.__qualname__ + instance_id
                    ]["last_results"]

            return wrapper

        return encapsulate

    def activate_communication(
        self, func: Union[str, Callable[..., Any]], mode: Union[str, List[str]]
    ):
        """
        Activates the communication mode for a registered function in the middleware communicator.
        The mode determines how the function will interact with the middleware communicator upon invocation,
        e.g., whether it should publish its return values, listen for values, etc.
        The communication mode can be set for all instances of the function or individually per instance.

        :param func: Union[str, Callable[..., Any]]: The function or the name of the function for which the communication mode should be activated
        :param mode: Union[str, List[str]]: Specifies the communication mode to be activated, either as a single mode (string) applicable to all instances, or as a list of modes (strings) per instance

        :raises: IndexError: If mode is a list and the number of elements does not match the number of instances
        :raises: ValueError: If the instance address cannot be found in the registry
        """
        if isinstance(func, str):
            func = getattr(self, func)
        entry = self.__registry.get(func.__qualname__, None)
        if entry is not None:
            instance_addr = hex(id(self))
            wrapyfi_instances = entry.get("__WRAPYFI_INSTANCES", None)
            if wrapyfi_instances is None:
                entry["__WRAPYFI_INSTANCES"] = [instance_addr]
                instance_id = 1
            else:
                try:
                    instance_id = wrapyfi_instances.index(instance_addr) + 1
                except ValueError:
                    wrapyfi_instances.append(instance_addr)
                    instance_id = len(wrapyfi_instances)
                    if instance_id > 1:
                        self.__registry[f"{func.__qualname__}.{instance_id}"] = (
                            deepcopy(entry, exclude_keys=["wrapped_executor"])
                        )

            instance_qualname = (
                f"{func.__qualname__}.{instance_id}"
                if instance_id > 1
                else func.__qualname__
            )

            if isinstance(mode, list):
                try:
                    self.__registry[instance_qualname]["mode"] = mode[instance_id - 1]
                except IndexError:
                    raise IndexError(
                        "When mode (publish|listen|disable|null) specified in configuration file is a "
                        "list, No. of elements in the list should match the number of instances"
                    )
            else:
                self.__registry[instance_qualname]["mode"] = mode
            self.__registry[instance_qualname]["instance_addr"] = instance_addr

    def close(self):
        """
        Closes this middleware communicator instance.
        """
        self.close_instance(hex(id(self)))

    @staticmethod
    def get_communicators():
        """
        Returns the available middleware communicators.

        :return: Set[str]: The available middleware communicators (excluding the fallback communicator)
        """
        return (pub.Publishers.mwares | lsn.Listeners.mwares) - {"fallback"}

    @classmethod
    def close_all_instances(cls):
        """
        Closes all instances of the middleware communicator.
        """
        close_instance_idxs = set()
        for val in cls._MiddlewareCommunicator__registry.values():
            instance_addr = val.get("instance_addr", False)
            close_instance_idxs.add(instance_addr) if instance_addr else None
        for close_instance_idx in close_instance_idxs:
            cls.close_instance(close_instance_idx)

    @classmethod
    def close_instance(cls, instance_addr: Optional[str] = None):
        """
        Closes a specific instance of the middleware communicator. Note that the instance address is the hexadecimal
        representation of the instance's id. If no instance address is provided, all instances will be closed.

        :param instance_addr: str: The unique identifier of the instance to be closed, defaults to None

        :raises: ValueError: If the instance address cannot be found in the registry
        """
        while True:
            del_entry = False
            del_entry_name = None
            del_entry_idx = None
            other_entry_keys = []
            # find self in registry and remove id from instances
            for entry_key, entry_val in cls._MiddlewareCommunicator__registry.items():
                if instance_addr in entry_val.get("__WRAPYFI_INSTANCES", []):
                    if not del_entry:
                        del_entry_idx = entry_val["__WRAPYFI_INSTANCES"].index(
                            instance_addr
                        )
                        if del_entry_idx == 0:
                            del_entry = entry_key
                        else:
                            del_entry = (
                                re.sub("\.\d+", "\.", entry_key)
                                + "."
                                + str(del_entry_idx + 1)
                            )
                        del_entry_name = re.sub("\.\d+", "\.", entry_key)
                    else:
                        if del_entry_name in entry_key:
                            other_entry_keys.append(entry_key)
                    entry_val["__WRAPYFI_INSTANCES"].remove(instance_addr)

            # delete registry entry and all its publishers/listeners/servers/clients
            if del_entry:
                for communicator in cls._MiddlewareCommunicator__registry[del_entry][
                    "communicator"
                ]:
                    wrapped_executor = communicator.get("wrapped_executor", False)
                    if wrapped_executor:
                        if isinstance(wrapped_executor, list):
                            for single_wrapped_executor in wrapped_executor:
                                single_wrapped_executor.close()
                        else:
                            wrapped_executor.close()
                        communicator.pop("wrapped_executor")
                if other_entry_keys:
                    del cls._MiddlewareCommunicator__registry[del_entry]
                else:
                    cls._MiddlewareCommunicator__registry[del_entry].pop(
                        "__WRAPYFI_INSTANCES"
                    )
                    cls._MiddlewareCommunicator__registry[del_entry].pop("mode")
                    cls._MiddlewareCommunicator__registry[del_entry].pop(
                        "instance_addr"
                    )

                # shift all entries backwards following the deleted one
                if del_entry_idx - 1 < len(other_entry_keys):
                    for other_entry_key in other_entry_keys[del_entry_idx:]:
                        new_key = re.split("\.(\d+)", other_entry_key)
                        if len(new_key) == 1:
                            new_key = new_key[0]
                        elif str(int(new_key[1]) - 1) == "1":
                            new_key = new_key[0]
                        else:
                            new_key = new_key[0] + "." + str(int(new_key[1]) - 1)
                        cls._MiddlewareCommunicator__registry[new_key] = (
                            cls._MiddlewareCommunicator__registry.pop(other_entry_key)
                        )
            else:
                break

    def __del__(self):
        self.close()



import argparse
import time

try:
    import numpy as np
    import pandas as pd
except ImportError:
    print("Install pandas and NumPy before running this script.")
try:
    import tensorflow as tf

    # avoid allocating all GPU memory assuming tf>=2.2
    gpus = tf.config.experimental.list_physical_devices("GPU")
    for gpu in gpus:
        tf.config.experimental.set_memory_growth(gpu, True)
except ImportError:
    tf = None

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


SHOULD_WAIT = True


class Benchmarker(MiddlewareCommunicator):

    @staticmethod
    def get_numpy_object(dims):
        return {"numpy": np.ones(dims)}

    @staticmethod
    def get_pandas_object(dims):
        return {
            "pandas": pd.DataFrame(
                np.ones(dims), index=None, columns=list(range(dims[-1]))
            )
        }

    @staticmethod
    def get_cupy_gpu_object(dims, gpu=0):
        import cupy as cp

        with cp.cuda.Device(gpu):
            cp_ones = cp.ones(dims, dtype=cp.float32)
        return {"cupy_gpu": cp_ones}

    @staticmethod
    def get_pyarrow_object(dims):
        import pyarrow as pa

        return {"pyarrow": pa.array(np.ones(dims).flatten())}

    @staticmethod
    def get_xarray_object(dims):
        import xarray as xr

        return {"xarray": xr.DataArray(np.ones(dims), name="example")}

    @staticmethod
    def get_dask_object(dims):
        import dask.array as da

        return {"dask": da.ones(dims)}

    @staticmethod
    def get_pillow_object(dims):
        from PIL import Image

        return {"pillow": Image.fromarray(np.ones(dims, dtype=np.uint8))}

    @staticmethod
    def get_tensorflow_object(dims):
        return {"tensorflow": tf.ones(dims)}

    @staticmethod
    def get_jax_object(dims):
        import jax as jx

        return {"jax": jx.numpy.ones(dims)}

    @staticmethod
    def get_mxnet_object(dims):
        import mxnet as mx

        return {"mxnet": mx.nd.ones(dims)}

    @staticmethod
    def get_mxnet_gpu_object(dims, gpu=0):
        import mxnet as mx

        return {"mxnet_gpu": mx.nd.ones(dims, ctx=mx.gpu(gpu))}

    @staticmethod
    def get_pytorch_object(dims):
        import torch as th

        return {"pytorch": th.ones(dims)}

    @staticmethod
    def get_pytorch_gpu_object(dims, gpu=0):
        import torch as th

        return {"pytorch_gpu": th.ones(dims, device=f"cuda:{gpu}")}

    @staticmethod
    def get_paddle_object(dims):
        import paddle as pa

        return {"paddle": pa.Tensor(pa.ones(dims), place=pa.CPUPlace())}

    @staticmethod
    def get_paddle_gpu_object(dims, gpu=0):
        import paddle as pa

        return {"paddle_gpu": pa.Tensor(pa.zeros(dims), place=pa.CUDAPlace(gpu))}

    def get_all_objects(self, count, plugin_name):
        obj = {"count": count, "timestamp": time.time()}
        object_creator = getattr(self, f"get_{plugin_name}_object")
        data_object = object_creator((args.height, args.width))
        if plugin_name == "dask":
            data_object["dask"] = data_object["dask"].compute()
        obj.update(data_object)
        return obj

    @MiddlewareCommunicator.register(
        "NativeObject",
        "yarp",
        "ExampleClass",
        "/example/get_native_objects",
        carrier="tcp",
        should_wait=SHOULD_WAIT,
    )
    def get_yarp_native_objects(self, count, plugin_name):
        return (self.get_all_objects(count, plugin_name),)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "ros",
        "ExampleClass",
        "/example/get_native_objects",
        carrier="tcp",
        should_wait=SHOULD_WAIT,
    )
    def get_ros_native_objects(self, count, plugin_name):
        return (self.get_all_objects(count, plugin_name),)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "ros2",
        "ExampleClass",
        "/example/get_native_objects",
        carrier="tcp",
        should_wait=SHOULD_WAIT,
    )
    def get_ros2_native_objects(self, count, plugin_name):
        return (self.get_all_objects(count, plugin_name),)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "zeromq",
        "ExampleClass",
        "/example/get_native_objects",
        carrier="tcp",
        should_wait=SHOULD_WAIT,
    )
    def get_zeromq_native_objects(self, count, plugin_name):
        return (self.get_all_objects(count, plugin_name),)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--publish",
        dest="mode",
        action="store_const",
        const="publish",
        default="listen",
        help="Publish mode",
    )
    parser.add_argument(
        "--listen",
        dest="mode",
        action="store_const",
        const="listen",
        default="listen",
        help="Listen mode (default)",
    )
    parser.add_argument(
        "--mwares",
        type=str,
        default=list(MiddlewareCommunicator.get_communicators()),
        choices=MiddlewareCommunicator.get_communicators(),
        nargs="+",
        help="The middlewares to use for transmission",
    )
    parser.add_argument(
        "--plugins",
        type=str,
        default=[
            "numpy",
            "pandas",
            "cupy_gpu",
            "pyarrow",
            "xarray",
            "dask",
            "tensorflow",
            "jax",
            "mxnet",
            "mxnet_gpu",
            "pytorch",
            "pytorch_gpu",
            "paddle",
            "paddle_gpu",
        ],
        nargs="+",
        help="The middlewares to use for transmission",
    )
    parser.add_argument(
        "--height", type=int, default=200, help="The tensor image height"
    )
    parser.add_argument("--width", type=int, default=200, help="The tensor image width")
    parser.add_argument(
        "--trials",
        type=int,
        default=2000,
        help="Number of trials to run per middleware",
    )
    parser.add_argument(
        "--skip-trials",
        type=int,
        default=0,
        help="Number of trials to skip before logging "
        "to csv to avoid warmup time logging",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    benchmarker = Benchmarker()
    benchmark_logger = pd.DataFrame(
        columns=["middleware", "plugin", "timestamp", "count", "delay"]
    )
    benchmark_iterator = {}

    for middleware_name in args.mwares:
        benchmark_iterator[middleware_name] = getattr(
            benchmarker, f"get_{middleware_name}_native_objects"
        )

    for middleware_name, method in benchmark_iterator.items():
        benchmarker.activate_communication(method, mode=args.mode)
        for plugin_name in args.plugins:
            time_acc_native_objects = []
            counter = -1
            while True:
                counter += 1
                (native_objects,) = method(counter, plugin_name)
                if native_objects is not None:
                    time_acc_native_objects.append(
                        time.time() - native_objects["timestamp"]
                    )
                    print(
                        f"{middleware_name} :: {plugin_name} :: delay:",
                        time_acc_native_objects[-1],
                        " Length:",
                        len(time_acc_native_objects),
                        " Count:",
                        native_objects["count"],
                    )
                    if args.trials - 1 == native_objects["count"]:
                        break
                    if counter > args.skip_trials:
                        new_row = pd.DataFrame(
                            {
                                "middleware": [middleware_name],
                                "plugin": [plugin_name],
                                "timestamp": [native_objects["timestamp"]],
                                "count": [native_objects["count"]],
                                "delay": [time_acc_native_objects[-1]],
                            }
                        )
                        benchmark_logger = pd.concat(
                            [benchmark_logger, new_row], ignore_index=True
                        )

                    if counter == 0:
                        if args.mode == "publish":
                            time.sleep(5)
                    else:
                        if args.mode == "publish":
                            time.sleep(0.1)

            time_acc_native_objects = pd.DataFrame(np.array(time_acc_native_objects))
            print(
                f"{middleware_name} :: {plugin_name} :: time statistics:",
                time_acc_native_objects.describe(),
            )
            time.sleep(5)
    benchmark_logger.to_csv(
        f"results/benchmarking_native_object_{args.mode}__{','.join(args.mwares)}__{','.join(args.plugins)}.csv",
        index=False,
    )



import os
from glob import glob
import threading
import importlib.util
import sys
from typing import Union, Callable, Any, Optional, List


WRAPYFI_PLUGIN_PATHS = "WRAPYFI_PLUGIN_PATHS"


lock = threading.Lock()


def deepcopy(
    obj: Any,
    exclude_keys: Optional[Union[list, tuple]] = None,
    shallow_keys: Optional[Union[list, tuple]] = None,
):
    """
    Deep copy an object, excluding specified keys.

    :param obj: Any: The object to deep copy
    :param exclude_keys: Union[list, tuple]: A list of keys to exclude from the deep copy
    :param shallow_keys: Union[list, tuple]: A list of keys to shallow copy
    """
    import copy

    if exclude_keys is None:
        return copy.deepcopy(obj)
    else:
        if isinstance(obj, list):
            return [deepcopy(item, exclude_keys) for item in obj]
        elif isinstance(obj, tuple):
            return tuple(deepcopy(item, exclude_keys) for item in obj)
        elif isinstance(obj, set):
            return {deepcopy(item, exclude_keys) for item in obj}
        elif isinstance(obj, dict):
            _shallows = shallow_keys or []
            ret = {
                key: deepcopy(val, exclude_keys)
                for key, val in obj.items()
                if key not in exclude_keys + _shallows
            }
            ret.update({key: val for key, val in obj.items() if key in _shallows})
            return ret
        else:
            return copy.deepcopy(obj)


def get_default_args(fnc: Callable[..., Any]):
    """
    Get the default arguments for a function.

    :param fnc: Callable[..., Any]: The function to get the default arguments for
    """
    import inspect

    signature = inspect.signature(fnc)
    return {
        k: v.default
        for k, v in signature.parameters.items()
        if v.default is not inspect.Parameter.empty
    }


def match_args(
    args: Union[list, tuple],
    kwargs: dict,
    src_args: Union[list, tuple],
    src_kwargs: dict,
):
    """
    Match and Substitute Arguments and Keyword Arguments using Specified Source Values.

    Navigate through the provided `args` and `kwargs`, identifying entries prefixed with "$" and substituting
    them with values from `src_args` and `src_kwargs` respectively, to dynamically modify the function call
    parameters using the source values.

    :param args: Union[list, tuple]: A list of arguments, potentially containing strings that indicate substitutable entries.
        Substitutable entries are prefixed with "$" and followed by either:
            - A digit (indicating an index to reference a value from `src_args`), or
            - Non-digit characters (indicating a key to reference a value from `src_kwargs`).
    :param kwargs: dict:
        A dictionary of keyword arguments, where values might be strings indicating substitutable entries,
        similar to the entries in `args`.
    :param src_args: Union[list, tuple]:
        A list of source arguments, intended to be referenced by substitutable entries within `args`.
    :param src_kwargs: dict:
        A dictionary of source keyword arguments, intended to be referenced by substitutable entries within `args`
        and `kwargs`.
    :return: Tuple[list, dict]: A tuple containing:
        - list: The new arguments, formed by substituting specified entries from `args` using `src_args` and `src_kwargs`.
        - dict: The new keyword arguments, formed by substituting specified entries from `kwargs` using `src_args` and `src_kwargs`.
    """
    new_args = []
    new_kwargs = {}
    for arg in args:
        if arg[0] == "$" and arg[1:].isdigit():
            new_args.append(src_args[int(arg[1:])])
        elif arg[0] == "$" and not arg[1:].isdigit():
            new_args.append(src_kwargs[arg[1:]])
        else:
            new_args.append(arg)

    for kwarg_key, kwarg_val in kwargs.items():
        if isinstance(kwarg_val, str) and "$" in kwarg_val and kwarg_val[1:].isdigit():
            new_kwargs[kwarg_key] = src_args[int(kwarg_val[1:])]
        elif (
            isinstance(kwarg_val, str)
            and "$" in kwarg_val
            and not kwarg_val[1:].isdigit()
        ):
            new_kwargs[kwarg_key] = src_kwargs[kwarg_val[1:]]
        else:
            new_kwargs[kwarg_key] = kwarg_val
    return tuple(new_args), new_kwargs


def dynamic_module_import(modules: List[str], globals: dict):
    """
    Dynamically import modules.

    :param modules: List[str]: A list of module names to import
    :param globals: dict: The globals dictionary to update
    """
    for module_name in modules:
        if not module_name.endswith(".py") or module_name.endswith("__.py"):
            continue
        module_name = module_name[:-3]
        module_name = module_name.replace("/", ".")
        try:
            module = __import__(module_name, fromlist=["*"])
        except ImportError as e:
            # print(module_name + " could not be imported.", e)
            continue
        if hasattr(module, "__all__"):
            all_names = module.__all__
        else:
            all_names = [name for name in dir(module) if not name.startswith("_")]
        globals.update({name: getattr(module, name) for name in all_names})


class SingletonOptimized(type):
    """
    A singleton metaclass that is thread-safe and optimized for speed.

    Source: https://stackoverflow.com/a/6798042
    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(SingletonOptimized, cls).__call__(
                        *args, **kwargs
                    )
        return cls._instances[cls]


class Plugin(object):
    """
    Base class for encoding and decoding plugins.
    """

    def encode(self, *args, **kwargs):
        """
        Encode data into a base64 string.

        :param args: tuple: Additional arguments
        :param kwargs: dict: Additional keyword arguments
        :return: Tuple[bool, dict]: A tuple containing:
            - bool: True if the encoding was successful, False otherwise
            - dict: A dictionary containing:
                - '__wrapyfi__': A tuple containing the class name and encoded data string
        """
        raise NotImplementedError

    def decode(self, *args, **kwargs):
        """
        Decode a base64 string back into data.

        :param args: tuple: Additional arguments
        :param kwargs: dict: Additional keyword arguments
        :return: Tuple[bool, object]: A tuple containing:
            - bool: True if the decoding was successful, False otherwise
            - object: The decoded data
        """
        raise NotImplementedError


class PluginRegistrar(object):
    """
    Class for registering encoding and decoding plugins.
    """

    encoder_registry = {}
    decoder_registry = {}

    @staticmethod
    def register(types=None):
        """
        Register a plugin for encoding and decoding a specific type.

        :param types: tuple: The type(s) to register the plugin for
        """

        def wrapper(cls):
            if types is not None:
                for cls_type in types:
                    PluginRegistrar.encoder_registry[cls_type] = cls
                PluginRegistrar.decoder_registry[str(cls.__name__)] = cls
            return cls

        return wrapper

    @staticmethod
    def scan():
        """
        Scan the plugins directory (Wrapyfi builtin and external) for plugins to register.
        This method is called automatically when the module is imported.
        """
        modules = glob(
            os.path.join(os.path.dirname(__file__), "plugins", "*.py"), recursive=True
        )
        modules = [
            "wrapyfi.plugins."
            + module.replace(os.path.dirname(__file__) + "/plugins/", "")
            for module in modules
        ]
        dynamic_module_import(modules, globals())

        extern_modules_paths = os.environ.get(WRAPYFI_PLUGIN_PATHS, "").split(":")
        for mod_group_idx, extern_module_path in enumerate(extern_modules_paths):
            extern_modules = glob(
                os.path.join(extern_module_path, "plugins", "*.py"), recursive=True
            )
            for mod_idx, extern_module in enumerate(extern_modules):
                spec = importlib.util.spec_from_file_location(
                    f"wrapyfi.extern{mod_group_idx}.plugins{mod_idx}", extern_module
                )
                module = importlib.util.module_from_spec(spec)
                sys.modules[spec.name] = module
                spec.loader.exec_module(module)
                extern_modules = [
                    f"wrapyfi.extern{mod_group_idx}.plugins{mod_idx}."
                    + extern_module.replace(extern_module_path + "/plugins/", "")
                ]
                dynamic_module_import(extern_modules, globals())



from wrapyfi.connect.listeners import Listeners
from wrapyfi.connect.publishers import Publishers
from wrapyfi.connect.servers import Servers
from wrapyfi.connect.clients import Clients

Listeners.scan()
Publishers.scan()
Servers.scan()
Clients.scan()



"""
A message requester and replier for native Python objects, and images using OpenCV.

This script demonstrates the capability to transmit native Python objects and images using
the MiddlewareCommunicator within the Wrapyfi library. The communication follows the REQ/REP pattern
allowing message requesting and replying functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject, Image and AudioChunk messages
    - Transmitting a Python object, an image, and audio chunk
    - Applying the REQ/REP pattern with persistence
    - Reading and transmitting an OpenCV image which can be loaded, resized, and displayed on the client and server ends
    - Reading and transmitting a sounddevice (PortAudio with NumPy) audio chunk which can be played back on the client and server ends

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)
    - OpenCV: Used for handling and creating images (installed with Wrapyfi)
    - sounddevice, soundfile: Used for handling audio

    Install using pip:
        ``pip install sounddevice soundfile``

Run:
    # Alternative 1: Image and NativeObject transmission
        # On machine 1 (or process 1): Requester sends a message and awaits a reply (image and native object)

        ``python3 request_reply_example.py --mode request --stream image``

        # On machine 2 (or process 2): Replier waits for a message and sends a reply (image and native object)

        ``python3 request_reply_example.py --mode reply --stream image``

    # Alternative 2: AudioChunk and NativeObject transmission
        # On machine 1 (or process 1): Requester sends a message and awaits a reply (audio chunk and native object)

        ``python3 request_reply_example.py --mode request --stream audio``

        # On machine 2 (or process 2): Replier waits for a message and sends a reply (audio chunk and native object)

        ``python3 request_reply_example.py --mode reply --stream audio``
"""

import argparse
import logging
import time

import sounddevice as sd
import soundfile as sf
import cv2
import numpy as np

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class ReqRep(MiddlewareCommunicator):

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "ReqRep",
        "/req_rep/my_message",
        carrier="tcp",
        persistent=True,
    )
    @MiddlewareCommunicator.register(
        "Image",
        "$mware",
        "ReqRep",
        "/req_rep/my_image_message",
        carrier="",
        width="$img_width",
        height="$img_height",
        rgb=True,
        jpg=True,
        persistent=True,
    )
    def send_img_message(
        self, msg=None, img_width=320, img_height=240, mware=None, *args, **kwargs
    ):
        """
        Exchange messages with OpenCV images and other native Python objects.
        """
        obj = {"message": msg, "args": args, "kwargs": kwargs}

        # read image from file
        img = cv2.imread("../../assets/wrapyfi.png")
        img = cv2.resize(img, (img_width, img_height), interpolation=cv2.INTER_AREA)
        cv2.putText(
            img,
            msg,
            (
                (
                    img.shape[1]
                    - cv2.getTextSize(msg, cv2.FONT_HERSHEY_SIMPLEX, 1, 2)[0][0]
                )
                // 2,
                (
                    img.shape[0]
                    + cv2.getTextSize(msg, cv2.FONT_HERSHEY_SIMPLEX, 1, 2)[0][1]
                )
                // 2,
            ),
            cv2.FONT_HERSHEY_SIMPLEX,
            1,
            (255, 0, 0),
            2,
            cv2.LINE_AA,
        )

        return obj, img

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware",
        "ReqRep",
        "/req_rep/my_message",
        carrier="tcp",
        persistent=True,
    )
    @MiddlewareCommunicator.register(
        "AudioChunk",
        "$mware",
        "ReqRep",
        "/req_rep/my_audio_message",
        carrier="",
        rate="$aud_rate",
        chunk="$aud_chunk",
        channels="$aud_channels",
        persistent=True,
    )
    def send_aud_message(
        self,
        msg=None,
        aud_rate=-1,
        aud_chunk=-1,
        aud_channels=2,
        mware=None,
        *args,
        **kwargs,
    ):
        """
        Exchange messages with sounddevice audio chunks and other native Python objects.
        """
        obj = {"message": msg, "args": args, "kwargs": kwargs}
        # read audio from file
        aud = sf.read("../../assets/sound_test.wav", dtype="float32")
        # aud = (np.mean(aud[0], axis=1), aud[1])
        return obj, aud


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(
        description="A message requester and replier for native Python objects, images using OpenCV, and sound using PortAudio."
    )
    parser.add_argument(
        "--mode",
        type=str,
        default="request",
        choices={"request", "reply"},
        help="The mode of communication, either 'request' or 'reply'",
    )
    parser.add_argument(
        "--mware",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission",
    )

    parser.add_argument(
        "--sound_device",
        type=int,
        default=0,
        help="The sound device to use for audio playback",
    )

    parser.add_argument(
        "--list_sound_devices",
        action="store_true",
        help="List all available sound devices and exit",
    )

    parser.add_argument(
        "--stream",
        type=str,
        default="image",
        choices={"image", "audio"},
        help="The streamed data as either 'image' or 'audio'",
    )
    return parser.parse_args()


def sound_play(my_aud, blocking=True, device=0):
    """
    Play audio using sounddevice.

    :param my_aud: Tuple[np.ndarray, int]: The audio chunk and sampling rate to play
    :param blocking: bool: Whether to block the execution until the audio is played
    :param device: int: The sound device to use for audio playback
    :return: bool: Whether the audio was played successfully
    """
    try:
        sd.play(*my_aud, blocking=blocking, device=device)
        return True
    except sd.PortAudioError:
        logging.warning(
            "PortAudioError: No device is found or the device is already in use. Will try again in 3 seconds."
        )
        return False


def main(args):
    if args.list_sound_devices:
        print(sd.query_devices())
        return

    """

    Main function to initiate ReqRep class and communication.
    """
    if args.mode == "request" and args.mware == "zeromq":
        print(
            "WE INTENTIONALLY WAIT 5 SECONDS TO ALLOW THE REPLIER ENOUGH TIME TO START UP. "
        )
        print(
            "THIS IS NEEDED WHEN USING ZEROMQ AS THE COMMUNICATION MIDDLEWARE IF THE SERVER "
        )
        print("IS SET TO SPAWN A PROXY BROKER (DEFAULT).")
        time.sleep(5)

    req_rep = ReqRep()
    if args.stream == "image":
        req_rep.activate_communication(ReqRep.send_img_message, mode=args.mode)
        req_rep.activate_communication(ReqRep.send_aud_message, mode="disable")
    elif args.stream == "audio":
        req_rep.activate_communication(ReqRep.send_aud_message, mode=args.mode)
        req_rep.activate_communication(ReqRep.send_img_message, mode="disable")
    counter = 0

    while True:
        # We separate the request and reply to show that messages are passed from the requester,
        # but this separation is NOT necessary for the method to work
        if args.mode == "request":
            msg = input("Type your message: ")
            my_message, my_image = req_rep.send_img_message(
                msg, counter=counter, mware=args.mware
            )
            (
                my_message2,
                my_aud,
            ) = req_rep.send_aud_message(msg, counter=counter, mware=args.mware)
            my_message = my_message2 if my_message2 is not None else my_message
            counter += 1
            if my_message is not None:
                print("Request: counter:", counter)
                print("Request: received reply:", my_message)
            if my_image is not None:
                cv2.imshow("Received image", my_image)
                while True:
                    k = cv2.waitKey(1) & 0xFF
                    if not (
                        cv2.getWindowProperty("Received image", cv2.WND_PROP_VISIBLE)
                    ):
                        break

                    if cv2.waitKey(1) == 27:
                        break  # esc to quit
                cv2.destroyAllWindows()
            if my_aud is not None:
                print("Request: received audio:", my_aud)
                while True:
                    played = sound_play(my_aud, blocking=True, device=args.sound_device)
                    if played:
                        break
                    else:
                        time.sleep(3)
        if args.mode == "reply":
            # The send_message() only executes in "reply" mode,
            # meaning, the method is only accessible from this code block
            my_message, my_image = req_rep.send_img_message(mware=args.mware)
            (
                my_message2,
                my_aud,
            ) = req_rep.send_aud_message(mware=args.mware)
            my_message = my_message2 if my_message2 is not None else my_message
            if my_message is not None:
                print("Reply: received reply:", my_message)
            if my_image is not None:
                cv2.imshow("Image", my_image)
                while True:
                    k = cv2.waitKey(1) & 0xFF
                    if not (cv2.getWindowProperty("Image", cv2.WND_PROP_VISIBLE)):
                        break

                    if cv2.waitKey(1) == 27:
                        break  # esc to quit
                cv2.destroyAllWindows()
            if my_aud is not None:
                while True:
                    played = sound_play(my_aud, blocking=True, device=args.sound_device)
                    if played:
                        break
                    else:
                        time.sleep(3)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
Forwarding Example using Wrapyfi.

This script demonstrates message forwarding using the MiddlewareCommunicator within the Wrapyfi library.
The communication follows chained forwarding through two methods, enabling PUB/SUB pattern
that allows message publishing and listening functionalities between processes or machines.

Demonstrations:
    - Using the NativeObject message
    - Transmitting a Python object with predefined message
    - Applying the PUB/SUB pattern with forwarding through two communication methods

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - YARP, ROS, ROS 2, ZeroMQ (refer to the Wrapyfi documentation for installation instructions)

Run:
    # On machine 1 (or process 1): Mode Chain A: Publisher [ZeroMQ], Chain B: Disabled (chain A & B can have different middleware
    by setting the --mware_... argument but it should be consistent on all machines or processes)
        ``python3 forwarding_example.py --mware_chain_A zeromq --mode_chain_A publish --mode_chain_B disable``

    # On machine 2 (or process 2): Mode Chain A: Listener [ZeroMQ], Chain B: Publisher [ROS 2] (chain A & B can have different middleware
    by setting the --mware_... argument but it should be consistent on all machines or processes)
        ``python3 forwarding_example.py --mware_chain_A zeromq --mode_chain_A listen --mware_chain_B ros2 --mode_chain_B publish``

    # On machine 3 (or process 3): Mode Chain A: Disabled, Chain B: Listener [ROS 2] (chain A & B can have different middleware
    by setting the --mware_... argument but it should be consistent on all machines or processes)
        ``python3 forwarding_example.py --mode_chain_A disable --mware_chain_B ros2 --mode_chain_B listen``
"""

import time
import argparse

from wrapyfi.connect.wrapper import MiddlewareCommunicator, DEFAULT_COMMUNICATOR


class ForwardCls(MiddlewareCommunicator):
    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware_chain_A",
        "ForwardCls",
        "/example/native_chain_A_msg",
        carrier="mcast",
        should_wait=True,
    )
    def read_chain_A(self, mware_chain_A=None, msg=""):
        """
        Read and forward message from chain A.
        """
        return (msg,)

    @MiddlewareCommunicator.register(
        "NativeObject",
        "$mware_chain_B",
        "ForwardCls",
        "/example/native_chain_B_msg",
        carrier="tcp",
        should_wait=False,
    )
    def read_chain_B(self, mware_chain_B=None, msg=""):
        """
        Read and forward message from chain B.
        """
        return (msg,)


def parse_args():
    """
    Parse command line arguments.
    """
    parser = argparse.ArgumentParser(description="Forwarding Example using Wrapyfi.")
    parser.add_argument(
        "--mode_chain_A",
        type=str,
        default="publish",
        choices=["listen", "publish", "disable", "none", None],
        help="The mode of transmission for the first method in the chain",
    )
    parser.add_argument(
        "--mode_chain_B",
        type=str,
        default="listen",
        choices=["listen", "publish", "disable", "none", None],
        help="The mode of transmission for the second method in the chain",
    )
    parser.add_argument(
        "--mware_chain_A",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission of the first method in the chain",
    )
    parser.add_argument(
        "--mware_chain_B",
        type=str,
        default=DEFAULT_COMMUNICATOR,
        choices=MiddlewareCommunicator.get_communicators(),
        help="The middleware to use for transmission of the second method in the chain",
    )
    return parser.parse_args()


def main(args):
    """
    Main function to initiate ForwardCls class and communication.
    """
    forward = ForwardCls()
    forward.activate_communication(forward.read_chain_A, mode=args.mode_chain_A)
    forward.activate_communication(forward.read_chain_B, mode=args.mode_chain_B)

    while True:
        (msg,) = forward.read_chain_A(
            mware_chain_A=args.mware_chain_A,
            msg=f"This argument message was sent from read_chain_A transmitted over "
            f"{args.mware_chain_A}",
        )
        if msg is not None:
            print(msg)
        (msg,) = forward.read_chain_B(
            mware_chain_B=args.mware_chain_B,
            msg=f"{msg}. It was then forwarded to read_chain_B over {args.mware_chain_B}",
        )
        if msg is not None:
            if args.mode_chain_B == "listen":
                print(
                    f"{msg}. This message is the last in the chain received over {args.mware_chain_B}"
                )
            else:
                print(msg)
        time.sleep(0.1)


if __name__ == "__main__":
    args = parse_args()
    main(args)



"""
This example shows how to use the MiddlewareCommunicator to send and receive messages over Websockets. It can be used to test the
functionality of the Websockets using the PUB/SUB pattern. The example can be run on a single
machine or on multiple machines. In this example (as with all other examples), the listener awaits a message over the topic '/hello/my_message'.

Requirements:
    - Wrapyfi: Middleware communication wrapper (refer to the Wrapyfi documentation for installation instructions)
    - Websockets (refer to the Wrapyfi documentation for installation instructions)

Run:
    # PUB/SUB mode - Listener waits for message and prints the received object (assuming the websocket server is running). Only one instance of the websocket_server.py should be running

        ``python3 listener_client.py``

"""

import argparse

from wrapyfi.connect.wrapper import MiddlewareCommunicator


class HelloWorld(MiddlewareCommunicator):

    @MiddlewareCommunicator.register(
        "NativeObject",
        "websocket",
        "HelloWorld",
        "/hello/my_message",
        carrier="tcp",
        should_wait=True,
    )
    def receive_message(self):
        """
        Exchange messages and mirror user input.
        """
        return (None,)


if __name__ == "__main__":
    hello_world = HelloWorld()
    hello_world.activate_communication(HelloWorld.receive_message, mode="listen")

    while True:
        (my_message,) = hello_world.receive_message()
        if my_message is not None:
            print("Method result:", my_message)



import argparse
import logging

logging.getLogger().setLevel(logging.INFO)

import zmq

from wrapyfi.middlewares.zeromq import ZeroMQMiddlewareReqRep, ZeroMQMiddlewarePubSub


def main(
    comm_type,
    socket_ip,
    socket_pub_port,
    socket_sub_port,
    socket_rep_port,
    socket_req_port,
    **kwargs,
):
    if comm_type == "pubsub":
        socket_pub_address = f"tcp://{socket_ip}:{socket_pub_port}"
        socket_sub_address = f"tcp://{socket_ip}:{socket_sub_port}"

        ZeroMQMiddlewarePubSub.activate(
            **{
                "socket_pub_address": socket_pub_address,
                "socket_sub_address": socket_sub_address,
                "proxy_broker_spawn": "process",
                "verbose": True,
            },
            **kwargs,
        )
        while True:
            pass
    elif comm_type == "reqrep":
        socket_rep_address = f"tcp://{socket_ip}:{socket_rep_port}"
        socket_req_address = f"tcp://{socket_ip}:{socket_req_port}"
        ZeroMQMiddlewareReqRep.activate(
            **{
                "socket_rep_address": socket_rep_address,
                "socket_req_address": socket_req_address,
                "proxy_broker_spawn": "process",
                "verbose": True,
            },
            **kwargs,
        )
        while True:
            pass

    elif comm_type == "pubsubpoll":  # debugging the xpub/xsub with a poller
        xpub_addr = f"tcp://{socket_ip}:{socket_pub_port}"
        xsub_addr = f"tcp://{socket_ip}:{socket_sub_port}"
        context = zmq.Context()
        xpub_socket = context.socket(zmq.XPUB)
        xpub_socket.bind(xpub_addr)
        xsub_socket = context.socket(zmq.XSUB)
        xsub_socket.bind(xsub_addr)

        poller = zmq.Poller()
        poller.register(xpub_socket, zmq.POLLIN)
        poller.register(xsub_socket, zmq.POLLIN)
        logging.info(f"[ZeroMQ] Intialising PUB/SUB device broker")
        while True:
            # get event
            event = dict(poller.poll(1000))
            if xpub_socket in event:
                message = xpub_socket.recv_multipart()
                # print("[ZeroMQ BROKER] xpub_socket recv message: %r" % message)
                xsub_socket.send_multipart(message)
            if xsub_socket in event:
                message = xsub_socket.recv_multipart()
                # print("[ZeroMQ BROKER] xsub_socket recv message: %r" % message)
                xpub_socket.send_multipart(message)


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--socket_ip", type=str, default="127.0.0.1", help="Socket IP address"
    )
    parser.add_argument(
        "--socket_pub_port", type=int, default=5555, help="Socket publishing port"
    )
    parser.add_argument(
        "--socket_sub_port", type=int, default=5556, help="Socket subscription port"
    )
    parser.add_argument(
        "--socket_rep_port", type=int, default=5559, help="Socket reply port"
    )
    parser.add_argument(
        "--socket_req_port", type=int, default=5560, help="Socket request port"
    )
    parser.add_argument(
        "--comm_type",
        type=str,
        default="pubsub",
        choices=["pubsub", "pubsubpoll", "reqrep"],
        help="The zeromq communication pattern",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    main(**vars(args))



import logging
import json
import time
import os
import threading
import numpy as np
import cv2
from typing import Optional, Tuple

from wrapyfi.connect.publishers import Publisher, Publishers, PublisherWatchDog
from wrapyfi.middlewares.websocket import WebSocketMiddlewarePubSub
from wrapyfi.encoders import JsonEncoder

SOCKET_IP = os.environ.get("WRAPYFI_WEBSOCKET_SOCKET_IP", "127.0.0.1")
SOCKET_PORT = int(os.environ.get("WRAPYFI_WEBSOCKET_SOCKET_PORT", 5000))
WEBSOCKET_NAMESPACE = os.environ.get("WRAPYFI_WEBSOCKET_NAMESPACE", "/")
WEBSOCKET_MONITOR_LISTENER_SPAWN = os.environ.get(
    "WRAPYFI_WEBSOCKET_MONITOR_LISTENER_SPAWN", "thread"
)
WATCHDOG_POLL_REPEAT = None

if WEBSOCKET_MONITOR_LISTENER_SPAWN == "process":
    WEBSOCKET_MONITOR_LISTENER_SPAWN = "thread"
    logging.warning(
        "[WebSocket] Wrapyfi does not support multiprocessing for Websockets. Please set "
        "the environment variable WRAPYFI_WEBSOCKET_MONITOR_LISTENER_SPAWN='thread'. "
        "Switching automatically to 'thread' mode."
    )


class WebSocketPublisher(Publisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        socket_ip: str = SOCKET_IP,
        socket_port: int = SOCKET_PORT,
        namespace: str = WEBSOCKET_NAMESPACE,
        monitor_listener_spawn: Optional[str] = WEBSOCKET_MONITOR_LISTENER_SPAWN,
        websocket_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        Initialize the publisher.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param socket_ip: str: IP address of the socket. Default is '127.0.0.1'
        :param socket_port: int: Port of the socket for publishing. Default is 5000
        :param namespace: str: Namespace of the WebSocket. Default is '/'
        :param monitor_listener_spawn: str: Whether to spawn the monitor listener as a process or thread. Default is 'thread'
        :param websocket_kwargs: dict: Additional kwargs for the WebSocket middleware
        :param kwargs: Additional kwargs for the publisher
        """
        super().__init__(name, out_topic, should_wait=should_wait, **kwargs)

        self.socket_address = f"http://{socket_ip}:{socket_port}{namespace}"

        WebSocketMiddlewarePubSub.activate(
            socket_address=self.socket_address,
            monitor_listener_spawn=monitor_listener_spawn,
            **(websocket_kwargs or {}),
        )

    def await_connection(
        self, out_topic: Optional[str] = None, repeats: Optional[int] = None
    ):
        """
        Wait for the connection to be established.

        :param out_topic: str: Name of the output topic
        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        if out_topic is None:
            out_topic = self.out_topic
        logging.info(f"[WebSocket] Waiting for output connection: {out_topic}")
        if repeats is None:
            repeats = -1 if self.should_wait else 0

        while repeats > 0 or repeats == -1:
            if repeats != -1:
                repeats -= 1
            connected = WebSocketMiddlewarePubSub._instance.is_connected()
            if connected:
                logging.info(f"[WebSocket] Output connection established: {out_topic}")
                return True
            time.sleep(0.02)
        return False

    def close(self):
        """
        Close the publisher.
        """
        logging.info(f"[WebSocket] Closing publisher for topic: {self.out_topic}")
        WebSocketMiddlewarePubSub._instance.socketio_client.disconnect()
        time.sleep(0.2)

    def __del__(self):
        self.close()


@Publishers.register("NativeObject", "websocket")
class WebSocketNativeObjectPublisher(WebSocketPublisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        multi_threaded: bool = False,
        serializer_kwargs: Optional[dict] = None,
        **kwargs,
    ):
        """
        The NativeObjectPublisher using the WebSocket message construct assuming a combination of python native objects
        and numpy arrays as input. Serializes the data (including plugins) using the encoder and sends it as a string.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate socket for each thread. Default is False
        :param serializer_kwargs: dict: Additional kwargs for the serializer
        :param kwargs: dict: Additional kwargs for the publisher
        """
        super().__init__(name, out_topic, should_wait=should_wait, **kwargs)
        if multi_threaded:
            self._thread_local_storage = threading.local()

        self._plugin_encoder = JsonEncoder
        self._plugin_kwargs = kwargs
        self._serializer_kwargs = serializer_kwargs or {}

        if not self.should_wait:
            PublisherWatchDog().add_publisher(self)

    def establish(self, repeats: Optional[int] = None, **kwargs):
        """
        Establish the connection to the publisher.

        :param repeats: int: Number of repeats to await connection. None for infinite. Default is None
        :return: bool: True if connection established, False otherwise
        """
        established = self.await_connection(repeats=repeats)
        return self.check_establishment(established)

    def publish(self, obj):
        """
        Publish the object to the middleware.

        :param obj: object: Object to publish
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        obj_str = json.dumps(
            obj,
            cls=self._plugin_encoder,
            **self._plugin_kwargs,
            **self._serializer_kwargs,
        )
        socketio_client = WebSocketMiddlewarePubSub._instance.socketio_client
        socketio_client.emit(self.out_topic, obj_str)


@Publishers.register("Image", "websocket")
class WebSocketImagePublisher(WebSocketNativeObjectPublisher):

    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        multi_threaded: bool = False,
        width: int = -1,
        height: int = -1,
        rgb: bool = True,
        fp: bool = False,
        jpg: bool = False,
        **kwargs,
    ):
        """
        The ImagePublisher using the WebSocket message construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate socket for each thread. Default is False
        :param width: int: Width of the image. Default is -1 meaning that the width is not fixed
        :param height: int: Height of the image. Default is -1 meaning that the height is not fixed
        :param rgb: bool: True if the image is RGB, False if it is grayscale. Default is True
        :param fp: bool: True if the image is floating point, False if it is integer. Default is False
        :param jpg: bool: True if the image should be compressed as JPG. Default is False
        """
        super().__init__(name, out_topic, should_wait=should_wait, **kwargs)
        self.width = width
        self.height = height
        self.rgb = rgb
        self.fp = fp
        self.jpg = jpg

        self._type = np.float32 if self.fp else np.uint8

    def publish(self, img: np.ndarray):
        """
        Publish the image to the middleware.

        :param img: np.ndarray: Image to publish formatted as a cv2 image np.ndarray[img_height, img_width, channels]
        """
        if img is None:
            return

        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        if (
            0 < self.width != img.shape[1]
            or 0 < self.height != img.shape[0]
            or not (
                (img.ndim == 2 and not self.rgb)
                or (img.ndim == 3 and self.rgb and img.shape[2] == 3)
            )
        ):
            raise ValueError("Incorrect image shape for publisher")
        if not img.flags["C_CONTIGUOUS"]:
            img = np.ascontiguousarray(img)

        socketio_client = WebSocketMiddlewarePubSub._instance.socketio_client

        if self.jpg:
            # Encode image as JPEG
            _, img_encoded = cv2.imencode(".jpg", img)
            img_bytes = img_encoded.tobytes()
            header = {"timestamp": time.time()}
        else:
            # Serialize numpy array to bytes
            img_bytes = img.tobytes()
            header = {
                "timestamp": time.time(),
                "shape": img.shape,
                "dtype": str(img.dtype),
            }

        # Emit the header and image bytes as separate items in a list
        socketio_client.emit(self.out_topic, [header, img_bytes])


@Publishers.register("AudioChunk", "websocket")
class WebSocketAudioChunkPublisher(WebSocketNativeObjectPublisher):
    def __init__(
        self,
        name: str,
        out_topic: str,
        should_wait: bool = True,
        multi_threaded: bool = False,
        channels: int = 1,
        rate: int = 44100,
        chunk: int = -1,
        **kwargs,
    ):
        """
        The AudioChunkPublisher using the WebSocket message construct assuming a numpy array as input.

        :param name: str: Name of the publisher
        :param out_topic: str: Name of the output topic (e.g. 'topic')
        :param should_wait: bool: Whether to wait for at least one listener before unblocking the script. Default is True
        :param multi_threaded: bool: Whether to use a separate socket for each thread. Default is False
        :param channels: int: Number of channels. Default is 1
        :param rate: int: Sampling rate. Default is 44100
        :param chunk: int: Chunk size. Default is -1 meaning that the chunk size is not fixed
        """
        super().__init__(
            name,
            out_topic,
            should_wait=should_wait,
            multi_threaded=multi_threaded,
            **kwargs,
        )
        self.channels = channels
        self.rate = rate
        self.chunk = chunk

    def publish(self, aud: Tuple[np.ndarray, int]):
        """
        Publish the audio chunk to the middleware.

        :param aud: Tuple[np.ndarray, int]: Audio chunk to publish formatted as (np.ndarray[audio_chunk, channels], int[samplerate])
        """
        if not self.established:
            established = self.establish(repeats=WATCHDOG_POLL_REPEAT)
            if not established:
                return
            else:
                time.sleep(0.2)

        aud_array, rate = aud
        if aud_array is None:
            return
        if 0 < self.rate != rate:
            raise ValueError("Incorrect audio rate for publisher")
        chunk, channels = (
            aud_array.shape if len(aud_array.shape) > 1 else (aud_array.shape[0], 1)
        )
        self.chunk = chunk if self.chunk == -1 else self.chunk
        self.channels = channels if self.channels == -1 else self.channels
        if 0 < self.chunk != chunk or 0 < self.channels != channels:
            raise ValueError("Incorrect audio shape for publisher")
        aud_array = np.require(aud_array, dtype=np.float32, requirements="C")

        socketio_client = WebSocketMiddlewarePubSub._instance.socketio_client

        aud_bytes = aud_array.tobytes()
        header = {
            "shape": aud_array.shape,
            "dtype": str(aud_array.dtype),
            "rate": rate,
            "timestamp": time.time(),
        }
        # Emit the header and audio bytes as a list
        socketio_client.emit(self.out_topic, [header, aud_bytes])


@Publishers.register("Properties", "websocket")
class WebSocketPropertiesPublisher(WebSocketPublisher):

    def __init__(self, name, out_topic, **kwargs):
        super().__init__(name, out_topic, **kwargs)
        raise NotImplementedError



import zmq
import json
import argparse
import logging

logging.getLogger().setLevel(logging.INFO)


def monitor_active_connections(socket_pub_address, topic):
    try:
        context = zmq.Context()
        subscriber = context.socket(zmq.SUB)

        subscriber.connect(socket_pub_address)
        subscriber.setsockopt_string(zmq.SUBSCRIBE, topic)

        while True:
            topic, message = subscriber.recv_multipart()
            topic = topic.decode("utf-8")
            data = json.loads(message.decode("utf-8"))
            logging.info(f"[ZeroMQ] Topic: {topic}, Data: {data}")

    except Exception as e:
        logging.error(f"[ZeroMQ] An error occurred: {str(e)}")


def parse_args():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--socket_pub_address",
        type=str,
        default="tcp://127.0.0.1:5555",
        help="Socket subscription address",
    )
    parser.add_argument(
        "--topic",
        type=str,
        default="ZEROMQ/CONNECTIONS",
        help="Topic to subscribe. The ZEROMQ/CONNECTIONS topic monitors subscribers to a any topic on "
        "the sub_address. Note that zeromq_proxy_broker.py in pubsub mode (either as a standalone"
        "or spawned by default by any publisher) must be running on socket_pub_address",
    )
    return parser.parse_args()


if __name__ == "__main__":
    args = parse_args()
    monitor_active_connections(args.socket_pub_address, args.topic)
